# Copyright (C) 2012 Rocco Aliberti
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import logging
import operator

logger = logging.getLogger(__name__)
import os
from urllib.parse import urlparse
import http.client
import socket

import xml.etree.ElementTree as ETree
from xl import event, main, playlist, xdg
from xl.radio import RadioStation, RadioList, RadioItem
from xl.nls import gettext as _
from xlgui.panel import radio

STATION = None


def enable(exaile):
    if exaile.loading:
        event.add_callback(_enable, 'exaile_loaded')
    else:
        _enable(None, exaile, None)


def _enable(o1, exaile, o2):
    global STATION

    STATION = SomaFMRadioStation()
    exaile.radio.add_station(STATION)


def disable(exaile):
    global STATION
    exaile.radio.remove_station(STATION)
    STATION = None


def set_status(message, timeout=0):
    radio.set_status(message, timeout)


class SomaFMRadioStation(RadioStation):

    name = "somafm"

    def __init__(self):
        """
        Initializes the somafm radio station
        """
        self.user_agent = main.exaile().get_user_agent_string('somafm')
        self.somafm_url = 'https://somafm.com/'
        self.channels_xml_url = self.somafm_url + 'channels.xml'
        self.cache_file = os.path.join(xdg.get_cache_dir(), 'somafm.cache')
        self.channelist = ''
        self.data = {}
        self._load_cache()
        self.subs = {}
        self.playlists = {}
        self.playlist_id = 0
        logger.debug(self.user_agent)

    def get_document(self, url):
        """
        Connects to the server and retrieves the document
        """
        set_status(_('Contacting SomaFM server...'))
        hostinfo = urlparse(url)

        try:
            c = http.client.HTTPConnection(hostinfo.netloc, timeout=20)
        except TypeError:
            c = http.client.HTTPConnection(hostinfo.netloc)

        try:
            c.request('GET', hostinfo.path, headers={'User-Agent': self.user_agent})
            response = c.getresponse()
        except (socket.timeout, socket.error):
            raise radio.RadioException(_('Error connecting to SomaFM server.'))

        if response.status != 200:
            raise radio.RadioException(_('Error connecting to SomaFM server.'))

        document = response.read()
        c.close()

        set_status('')
        return document

    def _load_cache(self):
        """
        Loads somafm data from cache
        """
        self.data = {}
        if os.path.isfile(self.cache_file):
            tree = ETree.parse(self.cache_file)
            for channel in tree.findall('channel'):
                self.data[channel.get("id")] = channel.get("name")

    def _save_cache(self):
        """
        Saves cache data
        """
        channellist = ETree.Element('channellist')
        for channel_id, channel_name in self.data.items():
            ETree.SubElement(channellist, 'channel', id=channel_id, name=channel_name)

        with open(self.cache_file, 'w') as h:
            h.write('<?xml version="1.0" encoding="UTF-8"?>')
            h.write(ETree.tostring(channellist, 'unicode'))

    def get_lists(self, no_cache=False):
        """
        Returns the rlists for somafm
        """
        if no_cache or not self.data:
            self.channellist = self.get_document(self.channels_xml_url)
            data = {}
            tree = ETree.fromstring(self.channellist)

            for channel in tree.findall('channel'):
                name = channel.find('title').text
                data[channel.get("id")] = name

            self.data = data
            self._save_cache()

        else:
            data = self.data

        rlists = []

        for id, name in data.items():
            rlist = RadioList(name, station=self)
            rlist.get_items = lambda no_cache, id=id: self._get_subrlists(
                id=id, no_cache=no_cache
            )
            rlists.append(rlist)

        rlists.sort(key=operator.attrgetter('name'))
        self.rlists = rlists

        return rlists

    def _get_subrlists(self, id, no_cache=False):
        """
        Gets the subrlists for a rlist
        """
        if no_cache or id not in self.subs:

            rlists = self._get_stations(id)
            rlists.sort(key=operator.attrgetter('name'))

            self.subs[id] = rlists

        return self.subs[id]

    def _get_playlist(self, url, playlist_id):
        """
        Gets the playlist for the given url and id
        """
        if playlist_id not in self.playlists:
            set_status(_('Contacting SomaFM server...'))
            try:
                self.playlists[playlist_id] = playlist.import_playlist(url)
            except Exception:
                set_status(_("Error importing playlist"))
                logger.exception("Error importing playlist")
            set_status('')

        return self.playlists[playlist_id]

    def _get_stations(self, id):
        if not self.channelist:
            self.channelist = self.get_document(self.channels_xml_url)

        tree = ETree.fromstring(self.channelist)
        channel = tree.find('.//channel[@id="%s"]' % id)
        plss = channel.findall('.//*[@format]')

        rlists = []
        for pls in plss:
            type = pls.tag.replace('pls', '')
            format = pls.attrib['format'].upper()
            url = pls.text
            display_name = format + " - " + type

            rlist = RadioItem(display_name, station=self)
            rlist.format = format
            rlist.get_playlist = (
                lambda url=url, playlist_id=self.playlist_id: self._get_playlist(
                    url, playlist_id
                )
            )

            self.playlist_id += 1
            rlists.append(rlist)
        return rlists

    def get_menu(self, parent):
        return parent.get_menu()



import collections
import os
import shutil
import tempfile

from gi.repository import Gio

import pytest

from xl.trax.track import Track

import logging

logging.basicConfig(level=logging.DEBUG)


@pytest.yield_fixture(autouse=True)
def exaile_test_cleanup():
    """
    Teardown/setup of various Exaile globals
    """

    yield

    Track._Track__the_cuts = ['the', 'a']

    Track._Track__tracksdict.clear()


#
# Fixtures for test track data
#


TrackData = collections.namedtuple(
    'TrackData',
    ['ext', 'filename', 'uri', 'size', 'writeable', 'has_cover', 'has_tags'],
)


def _fname(ext):
    local_path = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            'data',
            'music',
            'delerium',
            'chimera',
            '05 - Truly',
        )
        + os.extsep
        + ext
    )

    return ext, local_path, Gio.File.new_for_path(local_path).get_uri()


_all_tracks = [
    # fmt: off
    TrackData(*_fname('aac'),  size=9404,  writeable=True, has_cover=True, has_tags=True),
    TrackData(*_fname('aiff'), size=21340, writeable=True, has_cover=True, has_tags=True),
    TrackData(*_fname('au'),   size=16425, writeable=False, has_cover=False, has_tags=False),
    TrackData(*_fname('flac'), size=20668, writeable=True, has_cover=True, has_tags=True),
    TrackData(*_fname('mp3'),  size=7495,  writeable=True, has_cover=True, has_tags=True),
    TrackData(*_fname('mp4'),  size=7763,  writeable=True, has_cover=True, has_tags=True),
    TrackData(*_fname('mpc'),  size=6650,  writeable=True, has_cover=False, has_tags=True),
    TrackData(*_fname('ogg'),  size=17303, writeable=True, has_cover=True, has_tags=True),
    TrackData(*_fname('spx'),  size=1000,  writeable=True, has_cover=False, has_tags=True),
    TrackData(*_fname('wav'),  size=46124, writeable=False, has_cover=False, has_tags=False),
    TrackData(*_fname('wma'),  size=4929,  writeable=True, has_cover=False, has_tags=True),
    TrackData(*_fname('wv'),   size=32293, writeable=True, has_cover=False, has_tags=True),
    # fmt: on
]


_writeable_tracks = [t for t in _all_tracks if t.writeable]


@pytest.fixture(params=_all_tracks)
def test_track(request):
    '''Provides TrackData objects for each test track'''
    return request.param


@pytest.fixture(params=_writeable_tracks)
def writeable_track(request):
    '''Provides TrackData objects for each test track that is writeable'''
    return request.param


@pytest.yield_fixture()
def test_track_fp(test_track):
    with tempfile.NamedTemporaryFile(suffix='.' + test_track.ext) as tfp:
        with open(test_track.filename, 'rb') as fp:
            shutil.copyfileobj(fp, tfp)

        tfp.flush()

        yield tfp


@pytest.yield_fixture()
def writeable_track_name(writeable_track):
    '''Fixture that returns names of temporary copies of writeable tracks'''
    with tempfile.NamedTemporaryFile(suffix='.' + writeable_track.ext) as tfp:
        with open(writeable_track.filename, 'rb') as fp:
            shutil.copyfileobj(fp, tfp)

        tfp.flush()

        yield tfp.name


@pytest.fixture
def test_tracks():
    """
    Returns an object that can be used to retrieve test track data
    """

    class _TestTracks:
        def get(self, ext):
            return [x for x in _all_tracks if x.filename.endswith(ext)][0]

    return _TestTracks()



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import GLib
from gi.repository import Gtk

from xl.common import clamp, idle_add

from xlgui.guiutil import GtkTemplate


@GtkTemplate('ui', 'widgets', 'progress.ui')
class ProgressMonitor(Gtk.Box):
    """
    A graphical progress monitor designed to work with
    :class:`xl.common.ProgressThread`
    """

    __gtype_name__ = 'ProgressMonitor'

    label, progressbar = GtkTemplate.Child.widgets(2)

    def __init__(self, manager, thread, description, image=None):
        """
        Initializes the monitor

        :param manager: the parent manager
        :type manager: :class:`ProgressManager`
        :param thread: the thread to run
        :type thread: :class:`xl.common.ProgressThread`
        :param description: the description for this process
        :type description: string
        """
        super(ProgressMonitor, self).__init__()
        self.init_template()

        self.manager = manager
        self.thread = thread
        self._progress_updated = False

        if image is not None:
            self.pack_start(image, False, True, 0)

        self.label.set_text(description)

        self.show_all()
        GLib.timeout_add(100, self.pulsate_progress)

        self.progress_update_id = self.thread.connect(
            'progress-update', self.on_progress_update
        )
        self.done_id = self.thread.connect('done', self.on_done)
        self.thread.start()

    def destroy(self):
        """
        Cleans up
        """

        self._progress_updated = True

        if self.progress_update_id is not None:
            self.thread.disconnect(self.progress_update_id)
            self.thread.disconnect(self.done_id)

            self.progress_update_id = None
            self.done_id = None

    def pulsate_progress(self):
        """
        Pulses the progress indicator until
        the first status update is received
        """
        if self._progress_updated:
            return False

        self.progressbar.pulse()

        return True

    @idle_add()
    def on_progress_update(self, thread, progress):
        """
        Called when the progress has been updated
        """

        if progress is None:
            return

        # Accept a tuple or number between 0 and 100
        if hasattr(progress, '__len__'):
            step, total = progress
            percent = int(((step + 1) / total) * 100)
        else:
            percent = int(progress)

        if percent > 0:
            self._progress_updated = True

        fraction = clamp(percent / 100.0, 0, 1)

        self.progressbar.set_fraction(fraction)
        self.progressbar.set_text('%d%%' % percent)

    @idle_add()
    def on_done(self, thread):
        """
        Called when the thread is finished
        """
        self.manager.remove_monitor(self)

    @GtkTemplate.Callback
    def on_cancel_button_clicked(self, widget):
        """
        Stops the running thread
        """
        self.hide()
        self.thread.stop()
        self.destroy()


class ProgressManager:
    """
    Manages the [possibly multiple] progress bars that will allow the user
    to interact with different long running tasks that may occur in the
    application.

    The user should be able to see what task is running, the description,
    the current progress, and also be able to stop the task if they wish.
    """

    def __init__(self, container):
        """
        Initializes the manager

        :param container: the vertical Gtk.Box that will be holding the
        different progress indicators
        """
        self.box = container

    def add_monitor(self, thread, description, icon_name):
        """
        Adds a progress box

        :param thread: the ProgressThread that should be run once the
            monitor is started
        :param description: a description of the event
        :param icon_name: the name of an icon to display
        """
        image = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
        monitor = ProgressMonitor(self, thread, description, image)
        self.box.pack_start(monitor, False, True, 0)

        return monitor

    def remove_monitor(self, monitor):
        """
        Removes a monitor from the manager
        """
        monitor.hide()
        monitor.destroy()



# Matroska tagger for Exaile
# Copyright (C) 2010  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from xl.metadata import _base, _matroska


class MkaFormat(_base.BaseFormat):

    others = False  # For now, stick with defined tags only.
    writable = False

    tag_mapping = {
        'album': ('TITLE', 50),
        'album artist': ('ARTIST', 50),
        'artist': ('ARTIST', 30),
        'comment': ('COMMENT', 30),
        'composer': ('COMPOSER', 30),
        'date': ('DATE_RECORDED', 50),
        'discnumber': ('PART_NUMBER', 50),
        'genre': ('GENRE', 30),
        'performer': ('PERFORMER', 30),
        'title': ('TITLE', 30),
        'tracknumber': ('PART_NUMBER', 30),
    }

    def _get_raw(self):
        return self.tags

    def load(self):
        mka = _matroska.parse(self.loc)
        segment = mka['Segment'][0]
        info = segment['Info'][0]
        try:
            timecodescale = info['TimecodeScale'][0]
        except KeyError:
            timecodescale = 1000000
        length = info['Duration'][0] * timecodescale / 1e9
        self.tags = tags = {'__length': length}
        for mkatags in segment['Tags']:
            for mkatag in mkatags['Tag']:
                target = int(mkatag['Targets'][0]['TargetTypevalue'][0])
                for simpletag in mkatag['SimpleTag']:
                    key = (simpletag['TagName'][0], target)
                    try:
                        values = tags[key]
                    except KeyError:
                        values = tags[key] = []
                    values.append(simpletag['TagString'][0])


# vi: et sts=4 sw=4 ts=4



# Copyright (C) 2019  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import pickle
import shelve


class Utf8Unpickler(pickle.Unpickler):
    def __init__(self, *args, **kwargs):
        kwargs['encoding'] = 'utf-8'
        super().__init__(*args, **kwargs)


def ensure_shelve_compat():
    """"Change shelve's unpickler to use UTF-8, for compatibility with Python 2 shelves"""
    shelve.Unpickler = Utf8Unpickler



# Copyright (C) 2009-2010 Mathias Brodala
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gi.repository import Gtk

from xl import event, providers, settings
from xl.nls import gettext as _
from xlgui.accelerators import Accelerator
from xlgui.widgets import menu

from . import controls
from . import minimode_preferences

MINIMODE = None


def __migrate_fixed_controls():
    """
    Makes sure fixed controls are selected,
    mostly for migration from older versions
    """
    option_name = 'plugin/minimode/selected_controls'

    if settings.MANAGER.has_option(option_name):
        selected_controls = settings.get_option(option_name)

        if 'restore' not in selected_controls:
            selected_controls += ['restore']
            settings.set_option(option_name, selected_controls)


def enable(exaile):
    """
    Enables the mini mode plugin
    """
    __migrate_fixed_controls()

    if exaile.loading:
        event.add_callback(_enable, 'gui_loaded')
    else:
        _enable(None, exaile, None)


def _enable(event, exaile, nothing):
    """
    Handles the deferred enable call
    """
    global MINIMODE
    MINIMODE = MiniMode(exaile)


def disable(exaile):
    """
    Disables the mini mode plugin
    """
    global MINIMODE
    MINIMODE.destroy()
    MINIMODE = None


def get_preferences_pane():
    return minimode_preferences


class MiniMode(Gtk.Window):
    """
    Mini Mode main window
    """

    __gsignals__ = {'show': 'override'}

    def __init__(self, exaile):
        """
        Sets up the mini mode main window and
        options to access it
        """
        Gtk.Window.__init__(self)
        self.set_title('Exaile Mini Mode')
        self.set_resizable(False)

        self.exaile_window = exaile.gui.main.window

        controls.register()

        self.box = controls.ControlBox()
        self.box.set_spacing(3)
        self.border_frame = Gtk.Frame()
        self.border_frame.add(self.box)
        self.add(self.border_frame)

        self.accelerator = Accelerator(
            '<Primary><Alt>M', _('Mini Mode'), self.on_menuitem_activate
        )

        self.menuitem = menu.simple_menu_item(
            'minimode',
            ['clear-playlist'],
            icon_name='exaile-minimode',
            callback=self.accelerator,
        )

        providers.register('menubar-view-menu', self.menuitem)
        providers.register('mainwindow-accelerators', self.accelerator)

        self.mainbutton = Gtk.Button(label=_('Mini Mode'))
        self.mainbutton.set_image(
            Gtk.Image.new_from_icon_name('exaile-minimode', Gtk.IconSize.BUTTON)
        )
        self.mainbutton.connect('clicked', self.on_mainbutton_clicked)
        action_area = exaile.gui.main.info_area.get_action_area()
        action_area.pack_end(self.mainbutton, False, False, 6)

        self.__active = False
        self.__dirty = True
        # XXX: Until defaults are implemented in xl.settings
        self.__defaults = {
            'plugin/minimode/always_on_top': True,
            'plugin/minimode/show_in_panel': False,
            'plugin/minimode/on_all_desktops': True,
            'plugin/minimode/display_window_decorations': True,
            'plugin/minimode/window_decoration_type': 'full',
            'plugin/minimode/use_alpha': False,
            'plugin/minimode/transparency': 0.3,
            'plugin/minimode/horizontal_position': 10,
            'plugin/minimode/vertical_position': 10,
        }

        exaile.gui.main.connect('main-visible-toggle', self.on_main_visible_toggle)
        event.add_ui_callback(self.on_option_set, 'plugin_minimode_option_set')
        self.on_option_set(
            'plugin_minimode_option_set',
            settings,
            'plugin/minimode/button_in_mainwindow',
        )

    def destroy(self):
        """
        Cleanups
        """
        providers.unregister('mainwindow-accelerators', self.accelerator)
        providers.unregister('menubar-view-menu', self.menuitem)
        controls.unregister()

        self.mainbutton.destroy()

        self.set_active(False)
        self.box.destroy()
        Gtk.Window.destroy(self)

    def set_active(self, active):
        """
        Enables or disables the Mini Mode window
        """
        if active == self.__active:
            return

        if active and not self.props.visible:
            self.exaile_window.hide()
            self.show_all()
        elif not active and self.props.visible:
            self.hide()
            self.exaile_window.show()

        self.__active = active

    def do_show(self):
        """
        Updates the appearance if
        settings have been changed
        """
        h = None
        v = None

        if self.__dirty:
            for option, default in self.__defaults.items():
                value = settings.get_option(option, default)

                if option == 'plugin/minimode/always_on_top':
                    self.set_keep_above(value)
                elif option == 'plugin/minimode/show_in_panel':
                    self.props.skip_taskbar_hint = not value
                elif option == 'plugin/minimode/on_all_desktops':
                    if value:
                        self.stick()
                    else:
                        self.unstick()
                elif option == 'plugin/minimode/display_window_decorations':
                    if value:
                        option = 'plugin/minimode/window_decoration_type'
                        value = settings.get_option(option, self.__defaults[option])

                        if value == 'full':
                            self.set_decorated(True)
                            self.border_frame.set_shadow_type(Gtk.ShadowType.NONE)
                        elif value == 'simple':
                            self.set_decorated(False)
                            self.border_frame.set_shadow_type(Gtk.ShadowType.OUT)
                    else:
                        self.set_decorated(False)
                        self.border_frame.set_shadow_type(Gtk.ShadowType.NONE)
                elif option == 'plugin/minimode/use_alpha':
                    if value:
                        option = 'plugin/minimode/transparency'
                        opacity = 1 - settings.get_option(
                            option, self.__defaults[option]
                        )
                        self.set_opacity(opacity)
                elif option == 'plugin/minimode/horizontal_position':
                    h = value
                elif option == 'plugin/minimode/vertical_position':
                    v = value

            self.__dirty = False

        min_width, natural_width = self.get_preferred_width()
        min_height, natural_height = self.get_preferred_height()
        self.resize(natural_width, natural_height)
        self.queue_draw()
        Gtk.Window.do_show(self)

        # GTK (or perhaps the theme?) likes to move the window to some
        # random default position while showing it... so do these at the
        # same time after show, otherwise it'll move on us
        x, y = self.get_position()
        if h is not None:
            x = h
        if v is not None:
            y = v

        self.move(x, y)

    def do_configure_event(self, event):
        """
        Stores the window position upon window movement
        """
        settings.set_option('plugin/minimode/horizontal_position', event.x)
        settings.set_option('plugin/minimode/vertical_position', event.y)

    def do_delete_event(self, event):
        """
        Takes care of restoring Exaile's main window
        """
        self.set_active(False)

        return True

    def on_menuitem_activate(self, menuitem, name, parent, context):
        """
        Shows the Mini Mode window
        """
        self.set_active(True)

    def on_mainbutton_clicked(self, button):
        """
        Shows the Mini Mode window
        """
        self.set_active(True)

    def on_main_visible_toggle(self, main):
        """
        Handles visiblity toggles in
        Exaile's main window stead
        """
        if self.__active:
            if self.props.visible:
                self.hide()
            else:
                self.show_all()

            return True

        return False

    def on_option_set(self, event, settings, option):
        """
        Queues updates upon setting change
        """
        self.__dirty = True

        if option == 'plugin/minimode/button_in_mainwindow':
            button_in_mainwindow = settings.get_option(option, False)

            if button_in_mainwindow:
                self.mainbutton.set_no_show_all(False)
                self.mainbutton.show_all()
            else:
                self.mainbutton.hide()
                self.mainbutton.set_no_show_all(True)


# vim: et sts=4 sw=4



# Copyright (C) 2017 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import glob
import logging
import os
import shelve

from xl import common, shelve_compat

logger = logging.getLogger(__name__)


def migrate(path):
    shelve_compat.ensure_shelve_compat()

    bak_path = path + os.extsep + 'tmp'

    try:
        old_shelf = shelve.open(path, 'r', protocol=common.PICKLE_PROTOCOL)
    except Exception:
        logger.warning("%s may be corrupt", path)
        raise

    db = common.bsddb.hashopen(bak_path, 'c')
    new_shelf = shelve.BsdDbShelf(db, protocol=common.PICKLE_PROTOCOL)

    for k, v in old_shelf.items():
        new_shelf[k] = v

    new_shelf.close()
    old_shelf.close()

    try:
        os.replace(bak_path, path)
    except Exception:
        try:
            os.unlink(bak_path)
        except Exception:
            pass
        raise

    # Various types of *dbm modules use more than one file to store the data.
    # Now that we are assured that the migration is complete, delete them if
    # present
    for extra in glob.glob(path + os.extsep + '*'):
        try:
            os.unlink(extra)
        except Exception as e:
            logger.warning("Could not delete %s: %s", extra, e)

    logger.info("Migration successfully completed!")



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import importlib.util
import inspect
import logging
import os
import shutil
import sys
import tarfile

from xl.nls import gettext as _
from xl import event, settings, xdg

logger = logging.getLogger(__name__)


class InvalidPluginError(Exception):
    def __str__(self):
        return str(self.args[0])


class PluginsManager:
    def __init__(self, exaile, load=True):
        self.plugindirs = [os.path.join(p, 'plugins') for p in xdg.get_data_dirs()]
        if xdg.local_hack:
            self.plugindirs.insert(1, os.path.join(xdg.exaile_dir, 'plugins'))

        try:
            os.makedirs(self.plugindirs[0])
        except Exception:
            pass

        self.plugindirs = [x for x in self.plugindirs if os.path.exists(x)]
        self.loaded_plugins = {}

        self.exaile = exaile
        self.enabled_plugins = {}

        self.load = load

    def __findplugin(self, pluginname):
        for plugin_dir in self.plugindirs:
            path = os.path.join(plugin_dir, pluginname)
            if os.path.exists(path):
                return path
        return None

    def load_plugin(self, pluginname, reload_plugin=False):
        if not reload_plugin and pluginname in self.loaded_plugins:
            return self.loaded_plugins[pluginname]

        path = self.__findplugin(pluginname)
        if path is None:
            return False

        spec = importlib.util.spec_from_file_location(
            pluginname, os.path.join(path, '__init__.py')
        )
        plugin = importlib.util.module_from_spec(spec)
        # We need to temporarily add the plugin to sys.modules because otherwise the loader will fail to exec_module() if the plugin uses relative imports.
        if pluginname in sys.modules:
            raise InvalidPluginError(
                _('Plugin is already loaded or has a conflicting name.')
            )
        sys.modules[pluginname] = plugin
        spec.loader.exec_module(plugin)
        sys.modules[pluginname] = None

        if hasattr(plugin, 'plugin_class'):
            plugin = plugin.plugin_class()
        self.loaded_plugins[pluginname] = plugin
        return plugin

    def install_plugin(self, path):
        try:
            tar = tarfile.open(path, "r:*")  # transparently supports gz, bz2
        except (tarfile.ReadError, OSError):
            raise InvalidPluginError(_('Plugin archive is not in the correct format.'))

        # ensure the paths in the archive are sane
        mems = tar.getmembers()
        base = os.path.basename(path).split('.')[0]
        if os.path.isdir(os.path.join(self.plugindirs[0], base)):
            raise InvalidPluginError(
                _('A plugin with the name "%s" is already installed.') % base
            )

        for m in mems:
            if not m.name.startswith(base):
                raise InvalidPluginError(_('Plugin archive contains an unsafe path.'))

        tar.extractall(self.plugindirs[0])

    def __on_new_plugin_loaded(self, eventname, exaile, maybe_name, fn):
        event.remove_callback(self.__on_new_plugin_loaded, eventname)
        fn()

    def __enable_new_plugin(self, plugin):
        '''Sets up a new-style plugin. See helloworld plugin for details'''

        if hasattr(plugin, 'on_gui_loaded'):
            if self.exaile.loading:
                event.add_ui_callback(
                    self.__on_new_plugin_loaded,
                    'gui_loaded',
                    None,
                    plugin.on_gui_loaded,
                )
            else:
                plugin.on_gui_loaded()

        if hasattr(plugin, 'on_exaile_loaded'):
            if self.exaile.loading:
                event.add_ui_callback(
                    self.__on_new_plugin_loaded,
                    'exaile_loaded',
                    None,
                    plugin.on_exaile_loaded,
                )
            else:
                plugin.on_exaile_loaded()

    def uninstall_plugin(self, pluginname):
        self.disable_plugin(pluginname)
        for plugin_dir in self.plugindirs:
            try:
                shutil.rmtree(self.__findplugin(pluginname))
                return True
            except Exception:
                pass
        return False

    def enable_plugin(self, pluginname):
        try:
            plugin = self.load_plugin(pluginname)
            if not plugin:
                raise Exception("Error loading plugin")
            plugin.enable(self.exaile)
            if not inspect.ismodule(plugin):
                self.__enable_new_plugin(plugin)
            self.enabled_plugins[pluginname] = plugin
            logger.debug("Loaded plugin %s", pluginname)
            self.save_enabled()
            event.log_event('plugin_enabled', self, pluginname)
        except Exception as e:
            logger.exception("Unable to enable plugin %s", pluginname)
            raise e

    def disable_plugin(self, pluginname):
        try:
            plugin = self.enabled_plugins[pluginname]
            del self.enabled_plugins[pluginname]
        except KeyError:
            logger.exception("Plugin not found, possibly already disabled")
            return False
        try:
            plugin.disable(self.exaile)
            logger.debug("Unloaded plugin %s", pluginname)
            self.save_enabled()
        except Exception as e:
            logger.exception("Unable to fully disable plugin %s", pluginname)
            raise e
        finally:
            event.log_event('plugin_disabled', self, pluginname)
        return True

    def list_installed_plugins(self):
        pluginlist = []
        for directory in self.plugindirs:
            if not os.path.exists(directory):
                continue
            for name in os.listdir(directory):
                if (
                    name == '__pycache__'
                    or name in pluginlist
                    or not os.path.exists(os.path.join(directory, name, 'PLUGININFO'))
                ):
                    continue
                pluginlist.append(name)
        return pluginlist

    def list_available_plugins(self):
        pass

    def list_updateable_plugins(self):
        pass

    def get_plugin_info(self, pluginname):
        path = os.path.join(self.__findplugin(pluginname), 'PLUGININFO')
        infodict = {}
        with open(path) as f:
            for line in f:
                try:
                    key, val = line.split("=", 1)
                    # restricted eval - no bult-in funcs. marginally more secure.
                    infodict[key] = eval(val, {'__builtins__': None, '_': _}, {})
                except ValueError:
                    pass  # this happens on blank lines
        return infodict

    def is_compatible(self, info):
        """
        Returns True if the plugin claims to be compatible with the
        current platform.

        :param info: The data returned from get_plugin_info()
        """
        platforms = info.get('Platforms', [])
        if len(platforms) == 0:
            platforms = [sys.platform]

        for platform in platforms:
            if sys.platform.startswith(platform):
                return True

        return False

    def is_potentially_broken(self, info):
        """
        Returns True if one of the modules that the plugin requires is
        not detected as available.

        :param info: The data returned from get_plugin_info()
        """
        import pkgutil
        from gi.repository import GIRepository

        gir = GIRepository.Repository.get_default()

        modules = info.get('RequiredModules', [])

        for module in modules:
            pair = module.split(':', 1)
            if len(pair) > 1:
                prefix, module = pair
                if prefix == 'gi':
                    if not gir.enumerate_versions(module):
                        return True
            else:
                if not pkgutil.find_loader(module):
                    return True

        return False

    def save_enabled(self):
        if self.load:
            settings.set_option("plugins/enabled", list(self.enabled_plugins.keys()))

    def load_enabled(self):
        to_enable = settings.get_option("plugins/enabled", [])
        for plugin in to_enable:
            try:
                self.enable_plugin(plugin)
            except Exception:
                pass

    def teardown(self, main):
        """
        Tears down all enabled plugins
        """
        for plugin_name, plugin in self.enabled_plugins.items():
            if hasattr(plugin, 'teardown'):
                try:
                    plugin.teardown(main)
                except Exception:
                    logger.exception("Unable to tear down plugin %s", plugin_name)


# vim: et sts=4 sw=4



# Copyright (C) 2009-2010 Mathias Brodala
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

import logging

from xl import event, player, providers, settings
from xl.formatter import Formatter, TrackFormatter, ProgressTextFormatter
from xl.player.adapters import PlaybackAdapter, QueueAdapter
from xl.nls import gettext as _
from xlgui.guiutil import gtk_widget_replace
from xlgui.widgets.common import AttachedWindow
from xlgui.widgets.info import TrackToolTip
from xlgui.widgets.playback import SeekProgressBar
from xlgui.widgets.playlist import PlaylistModel, PlaylistView
from xlgui.widgets.rating import RatingWidget

logger = logging.getLogger(__name__)


def suppress(signal):
    """
    Decorator which prevents the emission of a GObject signal
    """

    def wrapper(function):
        def wrapped_function(self, *args, **kwargs):
            def on_event(sender, *args):
                sender.stop_emission_by_name(signal)
                return True

            handler_id = self.connect(signal, on_event)
            function(self, *args, **kwargs)
            self.disconnect(handler_id)

        return wrapped_function

    return wrapper


class ControlBox(Gtk.Box, providers.ProviderHandler):
    """
    A box for minimode controls which
    updates itself based on settings
    """

    __gsignals__ = {'show': 'override'}

    def __init__(self):
        Gtk.Box.__init__(self)
        providers.ProviderHandler.__init__(self, 'minimode-controls')

        self.__dirty = True
        self.__controls = {}

        event.add_ui_callback(self.on_option_set, 'plugin_minimode_option_set')

    def destroy(self):
        """
        Cleanups
        """
        for control in self.__controls.values():
            control.destroy()

    def __contains__(self, item):
        """
        Allows for checking for control ids
        """
        if item in self.__controls:
            return True

        return False

    def __getitem__(self, name):
        """
        Returns the control specified by name
        """
        return self.__controls[name]

    def __setitem__(self, name, control):
        """
        Sets the control specified by name
        """
        if name in self.__controls:
            self.remove(self.__controls[name])
            del self.__controls[name]

        self.__controls[name] = control
        self.pack_start(control, False, True, 0)
        control.show_all()

    def __delitem__(self, name):
        """
        Destroys the control specified by name
        """
        self.__controls[name].destroy()
        del self.__controls[name]

    def update(self):
        """
        Updates the controls to display
        """
        selected_controls = settings.get_option(
            'plugin/minimode/selected_controls',
            [
                'previous',
                'play_pause',
                'next',
                'playlist_button',
                'progress_bar',
                'restore',
            ],
        )

        added_controls = [c for c in selected_controls if c not in self]

        for name in added_controls:
            try:
                provider = self.get_provider(name)()
            except Exception:  # Not found, initialization error, ...)
                logger.exception('Failed to add control provider "%s"', name)
                selected_controls.remove(name)
            else:
                self[name] = provider

        removed_controls = [c.name for c in self if c.name not in selected_controls]

        for name in removed_controls:
            del self[name]

        for name in selected_controls:
            self.reorder_child(self[name], -1)

    def do_show(self):
        """
        Updates the appearance if
        settings have been changed
        """
        if self.__dirty:
            self.update()
            self.__dirty = False

        Gtk.Box.do_show(self)

    def on_provider_removed(self, provider):
        """
        Removes controls on provider removal
        """
        if provider.name in self:
            del self[provider.name]

    def on_option_set(self, event, settings, option):
        """
        Flags changes
        """
        if option == 'plugin/minimode/selected_controls':
            if self.props.visible:
                GLib.idle_add(self.update)
            else:
                self.__dirty = True


# Control definitions


class BaseControl:
    """
    Base control provider
    """

    name: str
    title: str
    description: str
    fixed = False


class ButtonControl(Gtk.Button, BaseControl):
    """
    Basic button control
    """

    __gsignals__ = {'clicked': 'override'}

    def __init__(self):
        Gtk.Button.__init__(self)
        BaseControl.__init__(self)

        self.set_image(Gtk.Image())
        self.set_relief(Gtk.ReliefStyle.NONE)
        self.set_focus_on_click(False)

    def set_image_from_icon_name(self, icon_name):
        """
        Sets the image to the specified icon

        :param icon_name:
        """
        self.props.image.set_from_icon_name(icon_name, Gtk.IconSize.BUTTON)


class PreviousButtonControl(ButtonControl):
    """
    Button which allows for going to the previous track
    """

    name = 'previous'
    title = _('Previous')
    description = _('Go to the previous track')

    def __init__(self):
        ButtonControl.__init__(self)

        self.set_image_from_icon_name('media-skip-backward')
        self.set_tooltip_text(_('Previous track'))

    def do_clicked(self):
        """
        Goes to the previous track
        """
        player.QUEUE.prev()


class NextButtonControl(ButtonControl):
    """
    Button which allows for going to the next track
    """

    name = 'next'
    title = _('Next')
    description = _('Go to the next track')

    def __init__(self):
        ButtonControl.__init__(self)

        self.set_image_from_icon_name('media-skip-forward')
        self.set_tooltip_text(_('Next track'))

    def do_clicked(self):
        """
        Goes to the next track
        """
        player.QUEUE.next()


class PlayPauseButtonControl(ButtonControl, PlaybackAdapter):
    """
    Button which allows for starting,
    pausing and resuming of playback
    """

    name = 'play_pause'
    title = _('Play/Pause')
    description = _('Start, pause or resume the playback')

    def __init__(self):
        ButtonControl.__init__(self)
        PlaybackAdapter.__init__(self, player.PLAYER)

        self.update_state()

    def destroy(self):
        """
        Cleanups
        """
        PlaybackAdapter.destroy(self)
        ButtonControl.destroy(self)

    def update_state(self):
        """
        Updates the appearance of this button
        """
        icon_name = 'media-playback-start'
        tooltip_text = _('Start playback')

        if not player.PLAYER.is_stopped():
            if player.PLAYER.is_paused():
                tooltip_text = _('Continue playback')
            elif player.PLAYER.is_playing():
                icon_name = 'media-playback-pause'
                tooltip_text = _('Pause playback')

        GLib.idle_add(self.set_image_from_icon_name, icon_name)
        GLib.idle_add(self.set_tooltip_text, tooltip_text)

    def do_clicked(self):
        """
        Starts, pauses or resumes the playback
        """
        if player.PLAYER.is_stopped():
            player.QUEUE.play()
        else:
            player.PLAYER.toggle_pause()

    def on_playback_track_start(self, event, player, track):
        """
        Updates state
        """
        self.update_state()

    def on_playback_player_end(self, event, player, track):
        """
        Updates state
        """
        self.update_state()

    def on_playback_toggle_pause(self, event, player, track):
        """
        Updates state
        """
        self.update_state()


class StopButtonControl(ButtonControl):
    """
    Button which allows for stopping the playback
    and toggling the SPAT feature
    """

    name = 'stop'
    title = _('Stop')
    description = _('Stop the playback')
    __gsignals__ = {
        'motion-notify-event': 'override',
        'leave-notify-event': 'override',
        'focus-out-event': 'override',
        'key-press-event': 'override',
        'key-release-event': 'override',
    }

    def __init__(self):
        ButtonControl.__init__(self)

        self.set_image_from_icon_name('media-playback-stop')
        self.set_tooltip_text(_('Stop playback'))

        self.add_events(
            Gdk.EventMask.LEAVE_NOTIFY_MASK
            | Gdk.EventMask.POINTER_MOTION_MASK
            | Gdk.EventMask.KEY_PRESS_MASK
            | Gdk.EventMask.KEY_RELEASE_MASK
        )

        self._queue_spat = False
        self._hovered = False

    def update_state(self):
        """
        Updates the appearance of this button
        """
        icon_name = 'media-playback-stop'
        tooltip_text = _('Stop playback')

        if self._queue_spat:
            icon_name = 'process-stop'

            if player.QUEUE.current_playlist.spat_position > 0:
                tooltip_text = _('Continue playback after current track')
            else:
                tooltip_text = _('Stop playback after current track')

        self.set_image_from_icon_name(icon_name)
        self.set_tooltip_text(tooltip_text)

    def do_clicked(self):
        """
        Stops the playback
        """
        if self._queue_spat:
            p = player.QUEUE.current_playlist
            p.spat_position = (
                -1 if p.current_position == p.spat_position else p.current_position
            )
        else:
            player.PLAYER.stop()

    def do_motion_notify_event(self, event):
        """
        Indicates SPAT
        """
        _, state = event.get_state()
        if state & Gdk.ModifierType.SHIFT_MASK:
            self._queue_spat = True
            self.update_state()

        self._hovered = True

    def do_leave_notify_event(self, event):
        """
        Returns to regular state
        """
        if self._queue_spat and not self.is_focus():
            self._queue_spat = False
            self.update_state()

        self._hovered = False

        ButtonControl.do_leave_notify_event(self, event)

    def do_focus_out_event(self, event):
        """
        Returns to regular state
        """
        if not self._hovered:
            self._queue_spat = False
            self.update_state()

    def do_key_press_event(self, event):
        """
        Indicates SPAT
        """
        if event.keyval in (Gdk.KEY_Shift_L, Gdk.KEY_Shift_R):
            self._queue_spat = True
            self.update_state()

    def do_key_release_event(self, event):
        """
        Returns to regular state
        """
        if event.keyval in (Gdk.KEY_Shift_L, Gdk.KEY_Shift_R):
            self._queue_spat = False
            self.update_state()


class VolumeButtonControl(Gtk.VolumeButton, BaseControl):
    """
    Button which allows for changing the volume
    """

    name = 'volume'
    title = _('Volume')
    description = _('Change the volume')
    __gsignals__ = {'value-changed': 'override'}

    def __init__(self):
        Gtk.VolumeButton.__init__(self)
        BaseControl.__init__(self)

        self.updating = False

        adjustment = Gtk.Adjustment(upper=1, step_incr=0.1, page_incr=0.2)
        self.set_adjustment(adjustment)

        # Slightly beautify the control buttons
        plus_button = self.get_plus_button()
        plus_button.set_image(
            Gtk.Image.new_from_icon_name('list-add', Gtk.IconSize.BUTTON)
        )
        plus_button.set_label('')
        minus_button = self.get_minus_button()
        minus_button.set_image(
            Gtk.Image.new_from_icon_name('list-remove', Gtk.IconSize.BUTTON)
        )
        minus_button.set_label('')

        event.add_ui_callback(self.on_option_set, 'player_option_set')
        self.on_option_set('player_option_set', settings, 'player/volume')

    def destroy(self):
        """
        Cleanups
        """
        event.remove_callback(self.on_option_set, 'player_option_set')

        ButtonControl.destroy(self)
        Gtk.VolumeButton.destroy(self)

    def set_value(self, value):
        """
        Override to take care of preventing
        signal handling and endless loops
        """
        self.updating = True
        Gtk.VolumeButton.set_value(self, value)
        self.updating = False

    def do_value_changed(self, value):
        """
        Changes the volume except if done internally
        """
        if not self.updating:
            settings.set_option('player/volume', value)

    def on_option_set(self, event, settings, option):
        """
        Reflects external volume changes
        """
        if option == 'player/volume':
            self.set_value(float(settings.get_option(option)))


class RestoreButtonControl(ButtonControl):
    """
    Button which allows for restoring the main window
    """

    name = 'restore'
    title = _('Restore')
    description = _('Restore the main window')
    fixed = True
    __gsignals__ = {'hierarchy-changed': 'override'}

    def __init__(self):
        ButtonControl.__init__(self)

        self.set_image_from_icon_name('window-new')
        self.set_tooltip_text(_('Restore main window'))

    def do_clicked(self):
        """
        Restores the main window
        """
        self.get_toplevel().set_active(False)

    def do_hierarchy_changed(self, previous_toplevel):
        """
        Sets up accelerators
        """
        accel_group = Gtk.AccelGroup()

        try:
            self.get_toplevel().add_accel_group(accel_group)
        except AttributeError:
            pass
        else:
            key, modifier = Gtk.accelerator_parse('<Primary><Alt>M')
            self.add_accelerator(
                'clicked', accel_group, key, modifier, Gtk.AccelFlags.VISIBLE
            )


class RatingControl(RatingWidget, BaseControl):
    """
    Control which allows for viewing and
    changing the rating of the current track
    """

    name = 'rating'
    title = _('Rating')
    description = _('Select rating of the current track')

    def __init__(self):
        RatingWidget.__init__(self, player=player.PLAYER)
        BaseControl.__init__(self)

    def do_rating_changed(self, rating):
        """
        Updates the rating of the currently playing track
        """
        if player.PLAYER.current is not None:
            player.PLAYER.current.set_rating(rating)
            maximum = settings.get_option('rating/maximum', 5)
            event.log_event('rating_changed', self, 100 * rating / maximum)


class TrackSelectorControl(Gtk.ComboBox, BaseControl, QueueAdapter):
    name = 'track_selector'
    title = _('Track selector')
    description = _('Simple track list selector')
    __gsignals__ = {'changed': 'override'}

    def __init__(self):
        Gtk.ComboBox.__init__(self)
        BaseControl.__init__(self)
        QueueAdapter.__init__(self, player.QUEUE)

        self.formatter = TrackFormatter('')
        self.model = Gtk.ListStore(object)
        self.set_model(self.model)

        self.synchronize()

        renderer = Gtk.CellRendererText()
        self.pack_start(renderer, True)
        self.set_cell_data_func(renderer, self.data_func)
        self.set_size_request(200, 0)

        event.add_ui_callback(self.on_option_set, 'plugin_minimode_option_set')
        self.on_option_set(
            'plugin_minimode_option_set', settings, 'plugin/minimode/track_title_format'
        )

    def destroy(self):
        """
        Cleanups
        """
        QueueAdapter.destroy(self)
        Gtk.ComboBox.destroy(self)

    def data_func(self, column, cell, model, iter, user_data):
        """
        Updates track titles and highlights
        the current track if the popup is shown
        """
        track = model.get_value(iter, 0)

        if track is None:
            return

        cell.props.text = self.formatter.format(track)

        active_iter = self.get_active_iter()

        if active_iter is not None:
            active_track = model.get_value(active_iter, 0)
            weight = Pango.Weight.NORMAL

            if self.props.popup_shown and track == active_track:
                weight = Pango.Weight.BOLD

            cell.props.weight = weight

    @suppress('changed')
    def synchronize(self):
        """
        Synchronizes the model data with
        the current content of the queue
        """
        self.set_model(None)
        self.model.clear()

        for i, track in enumerate(player.QUEUE.current_playlist):
            self.model.append([track])

            if track is player.QUEUE.current_playlist.current:
                # Not using iter since model is detached
                self.set_active(i)

        self.set_model(self.model)

    def do_changed(self):
        """
        Starts playing the selected track. Should only be
        triggered by user action to prevent race conditions
        """
        active_index = self.get_active()

        if active_index > -1:
            player.QUEUE.current_playlist.current_position = active_index
            player.QUEUE.play(player.QUEUE.current_playlist[active_index])

    def add_tracks(self, tracks):
        """
        Adds tracks to the internal storage
        """
        if not tracks:
            return

        self.set_model(None)

        for position, track in tracks:
            self.model.insert(position, [track])

        self.set_model(self.model)

    def remove_tracks(self, tracks):
        """
        Removes tracks from the internal storage
        """
        if not tracks:
            return

        self.set_model(None)
        tracks.reverse()

        for position, track in tracks:
            del self.model[position]

        self.set_model(self.model)

    def on_queue_current_playlist_changed(self, event, queue, playlist):
        """
        Updates the list on queue changes
        """
        self.synchronize()

    @suppress('changed')
    def on_queue_current_position_changed(self, event, playlist, positions):
        """
        Updates the list on queue changes
        """
        if positions[0] < 0:
            return

        GLib.idle_add(self.set_active, positions[0])

    def on_queue_tracks_added(self, event, queue, tracks):
        """
        Updates the list on queue changes
        """
        GLib.idle_add(self.add_tracks, tracks)

    def on_queue_tracks_removed(self, event, queue, tracks):
        """
        Updates the list on queue changes
        """
        GLib.idle_add(self.remove_tracks, tracks)

    def on_option_set(self, event, settings, option):
        """
        Updates control upon setting change
        """
        if option == 'plugin/minimode/track_title_format':
            self.formatter.set_property(
                'format', settings.get_option(option, _('$tracknumber - $title'))
            )


class PlaylistButtonControl(Gtk.ToggleButton, BaseControl, QueueAdapter):
    name = 'playlist_button'
    title = _('Playlist button')
    description = _('Access the current playlist')
    __gsignals__ = {'scroll-event': 'override'}

    def __init__(self):
        Gtk.ToggleButton.__init__(self)
        BaseControl.__init__(self)
        QueueAdapter.__init__(self, player.QUEUE)

        self.set_focus_on_click(False)
        self.set_size_request(200, -1)
        box = Gtk.Box()
        self.arrow = Gtk.Arrow(Gtk.ArrowType.RIGHT, Gtk.ShadowType.OUT)
        box.pack_start(self.arrow, False, True, 0)
        self.label = Gtk.Label(label='')
        self.label.props.ellipsize = Pango.EllipsizeMode.END
        box.pack_start(self.label, True, True, 0)
        self.add(box)

        self.formatter = TrackFormatter(
            settings.get_option(
                'plugin/minimode/track_title_format', '$tracknumber - $title'
            )
        )

        self.view = PlaylistView(player.QUEUE.current_playlist, player.PLAYER)
        self.popup = AttachedWindow(self)
        self.popup.set_default_size(
            settings.get_option('plugin/minimode/' 'playlist_button_popup_width', 350),
            settings.get_option('plugin/minimode/' 'playlist_button_popup_height', 400),
        )
        scrollwindow = Gtk.ScrolledWindow()
        scrollwindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrollwindow.set_shadow_type(Gtk.ShadowType.IN)
        scrollwindow.add(self.view)
        self.popup.add(scrollwindow)
        self.popup.connect('show', self.on_popup_show)
        self.popup.connect('hide', self.on_popup_hide)
        self.popup.connect('configure-event', self.on_popup_configure_event)

        accel_group = Gtk.AccelGroup()
        key, modifier = Gtk.accelerator_parse('<Primary>J')
        accel_group.connect(
            key, modifier, Gtk.AccelFlags.VISIBLE, self.on_accelerator_activate
        )
        self.popup.add_accel_group(accel_group)

        self.tooltip = TrackToolTip(self, player.PLAYER)
        self.tooltip.set_auto_update(True)

        if player.PLAYER.current is not None:
            self.label.set_text(self.formatter.format(player.PLAYER.current))

        self._drag_motion_timeout_id = None
        self._drag_leave_timeout_id = None

        self.drag_dest_set(
            Gtk.DestDefaults.ALL,
            self.view.targets,
            Gdk.DragAction.COPY | Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE,
        )

        self.connect('drag-motion', self.on_drag_motion)
        self.connect('drag-leave', self.on_drag_leave)
        self.connect('drag-data-received', self.on_drag_data_received)
        self.view.connect('drag-motion', self.on_drag_motion)
        self.view.connect('drag-leave', self.on_drag_leave)
        event.add_ui_callback(self.on_track_tags_changed, 'track_tags_changed')
        event.add_ui_callback(self.on_option_set, 'plugin_minimode_option_set')
        self.on_option_set(
            'plugin_minimode_option_set', settings, 'plugin/minimode/track_title_format'
        )

    def destroy(self):
        """
        Cleanups
        """
        self.tooltip.destroy()
        QueueAdapter.destroy(self)
        Gtk.ToggleButton.destroy(self)

    def update_playlist(self, playlist):
        """
        Updates the internally stored playlist
        """
        columns = self.view.model.column_names
        model = PlaylistModel(playlist, columns, player.PLAYER, self.view)
        self.view.set_model(model)

    def do_scroll_event(self, event):
        """
        Changes the current track
        """
        if event.direction == Gdk.ScrollDirection.UP:
            self.view.playlist.prev()
        elif event.direction == Gdk.ScrollDirection.DOWN:
            self.view.playlist.next()
        else:
            return

        position = self.view.playlist.current_position

        try:
            track = self.view.playlist[position]
        except IndexError:
            pass
        else:
            player.QUEUE.play(track)

    def do_toggled(self):
        """
        Shows or hides the playlist
        """
        if self.get_active():
            self.arrow.props.arrow_type = Gtk.ArrowType.DOWN
            self.popup.show_all()
        else:
            self.popup.hide()
            self.arrow.props.arrow_type = Gtk.ArrowType.RIGHT

    def on_accelerator_activate(self, accel_group, acceleratable, keyval, modifier):
        """
        Shows the current track
        """
        self.view.scroll_to_cell(self.view.playlist.current_position)
        self.view.set_cursor(self.view.playlist.current_position)

    def on_drag_motion(self, widget, context, x, y, time):
        """
        Prepares to show the playlist
        """
        # Defer display of the playlist
        if self._drag_motion_timeout_id is None:
            self._drag_motion_timeout_id = GLib.timeout_add(
                500, lambda: self.set_active(True)
            )

        # Prevent hiding of the playlist
        if self._drag_leave_timeout_id is not None:
            GLib.source_remove(self._drag_leave_timeout_id)
            self._drag_leave_timeout_id = None

    def on_drag_leave(self, widget, context, time):
        """
        Prepares to hide the playlist
        """
        # Enable display of the playlist on re-enter
        if self._drag_motion_timeout_id is not None:
            GLib.source_remove(self._drag_motion_timeout_id)
            self._drag_motion_timeout_id = None

        if self._drag_leave_timeout_id is not None:
            GLib.source_remove(self._drag_leave_timeout_id)

        # Defer hiding of the playlist
        self._drag_leave_timeout_id = GLib.timeout_add(
            500, lambda: self.set_active(False)
        )

    def on_drag_data_received(self, widget, context, x, y, selection, info, time):
        """
        Handles dropped data
        """
        # Enable display of the playlist on re-enter
        if self._drag_motion_timeout_id is not None:
            GLib.source_remove(self._drag_motion_timeout_id)
            self._drag_motion_timeout_id = None

        # Enable hiding of the playlist on re-enter
        if self._drag_leave_timeout_id is not None:
            GLib.source_remove(self._drag_leave_timeout_id)
            self._drag_leave_timeout_id = None

        self.view.emit('drag-data-received', context, x, y, selection, info, time)

    def on_popup_show(self, widget):
        if not self.get_active():
            self.set_active(True)

    def on_popup_hide(self, widget):
        if self.get_active():
            self.set_active(False)

    def on_popup_configure_event(self, widget, event):
        """
        Saves the window size after resizing
        """
        width = settings.get_option(
            'plugin/minimode/' 'playlist_button_popup_width', 350
        )
        height = settings.get_option(
            'plugin/minimode/' 'playlist_button_popup_height', 400
        )

        if event.width != width:
            settings.set_option(
                'plugin/minimode/' 'playlist_button_popup_width', event.width
            )

        if event.height != height:
            settings.set_option(
                'plugin/minimode/' 'playlist_button_popup_height', event.height
            )

    def on_queue_current_playlist_changed(self, event, queue, playlist):
        """
        Updates the list on queue changes
        """
        GLib.idle_add(self.update_playlist, playlist)

    def on_queue_current_position_changed(self, event, playlist, positions):
        """
        Updates the list on queue changes
        """
        try:
            track = playlist[positions[0]]
        except IndexError:
            text = ''
        else:
            text = self.formatter.format(track)

        GLib.idle_add(self.label.set_text, text)

    def on_track_tags_changed(self, event, track, tags):
        """
        Updates the button on tag changes
        """
        playlist = self.view.playlist

        if track not in playlist:
            return

        track_position = playlist.index(track)
        if track_position == playlist.current_position:
            self.label.set_text(self.formatter.format(track))

    def on_option_set(self, event, settings, option):
        """
        Updates control upon setting change
        """
        if option == 'plugin/minimode/track_title_format':
            self.formatter.set_property(
                'format', settings.get_option(option, _('$tracknumber - $title'))
            )


class ProgressButtonFormatter(Formatter):
    """
    Formatter which allows both for display of
    tag data as well as progress information
    """

    def __init__(self):
        Formatter.__init__(self, self.get_option_value())

        self.track_formatter = TrackFormatter('')
        self.progress_formatter = ProgressTextFormatter(
            self.props.format, player.PLAYER
        )

        event.add_ui_callback(self.on_option_set, 'plugin_minimode_option_set')

    def format(self, current_time=None, total_time=None):
        """
        Returns a string suitable for progress buttons

        :param current_time: the current progress
        :type current_time: float
        :param total_time: the total length of a track
        :type total_time: float
        :returns: The formatted text
        :rtype: string
        """
        text = self.progress_formatter.format()
        self.track_formatter.props.format = text
        text = self.track_formatter.format(player.PLAYER.current)

        return text

    def get_option_value(self):
        """
        Retrieves the current user format
        """
        return settings.get_option(
            'plugin/minimode/progress_button_title_format',
            _('$title ($current_time / $total_time)'),
        )

    def on_option_set(self, event, settings, option):
        """
        Updates the internal format on setting change
        """
        if option == 'gui/progress_bar_text_format':
            GLib.idle_add(self.set_property, 'format', self.get_option_value())


Gtk.rc_parse_string(
    '''
    style "progress-button" {
        GtkToggleButton::default-border = {0, 0, 0, 0}
        GtkToggleButton::default-outside-border = {0, 0, 0, 0}
        GtkToggleButton::inner-border = {0, 0, 0, 0}
    }
    widget "*.progressbutton" style "progress-button"
'''
)


class ProgressButtonControl(PlaylistButtonControl):
    name = 'progress_button'
    title = _('Progress button')
    description = _('Playback progress and access ' 'to the current playlist')
    # Required to make overrides work
    __gsignals__ = {}

    def __init__(self):
        PlaylistButtonControl.__init__(self)

        self.set_name('progressbutton')
        self.add_events(Gdk.EventMask.POINTER_MOTION_MASK)

        self.progressbar = SeekProgressBar(player.PLAYER)
        self.progressbar.set_size_request(-1, 1)
        self.progressbar.formatter = ProgressButtonFormatter()
        self.progressbar.set_text = lambda *a: None  # Needed by PlaylistButtonControl
        gtk_widget_replace(self.label, self.progressbar)
        self.label = self.progressbar

        if player.PLAYER.current is not None:
            self.progressbar.on_playback_track_start(
                'playback_track_start', player.PLAYER, player.PLAYER.current
            )

        self.tooltip = TrackToolTip(self, player.PLAYER)
        self.tooltip.set_auto_update(True)

    def destroy(self):
        """
        Cleanups
        """
        self.tooltip.destroy()
        PlaylistButtonControl.destroy(self)

    def do_button_press_event(self, event):
        """
        Trigger normal toggle action or seek
        """
        if event.button == Gdk.BUTTON_PRIMARY:
            PlaylistButtonControl.do_button_press_event(self, event)
        elif event.button == Gdk.BUTTON_MIDDLE:
            event = event.copy()
            event.button = Gdk.BUTTON_PRIMARY
            x, y = self.translate_coordinates(
                self.progressbar, int(event.x), int(event.y)
            )
            event.x, event.y = float(x), float(y)
            self.progressbar.emit('button-press-event', event)

    def do_button_release_event(self, event):
        if event.button == Gdk.BUTTON_PRIMARY:
            PlaylistButtonControl.do_button_release_event(self, event)
        elif event.button == Gdk.BUTTON_MIDDLE:
            event = event.copy()
            event.button = Gdk.BUTTON_PRIMARY
            x, y = self.translate_coordinates(
                self.progressbar, int(event.x), int(event.y)
            )
            event.x, event.y = float(x), float(y)
            self.progressbar.emit('button-release-event', event)

    def do_motion_notify_event(self, event):
        event = event.copy()
        x, y = self.translate_coordinates(self.progressbar, int(event.x), int(event.y))
        event.x, event.y = float(x), float(y)
        self.progressbar.emit('motion-notify-event', event)

    def do_leave_notify_event(self, event):
        event = event.copy()
        x, y = self.translate_coordinates(self.progressbar, int(event.x), int(event.y))
        event.x, event.y = float(x), float(y)
        self.progressbar.emit('leave-notify-event', event)


class ProgressBarControl(SeekProgressBar, BaseControl):
    name = 'progress_bar'
    title = _('Progress bar')
    description = _('Playback progress and seeking')

    def __init__(self):
        SeekProgressBar.__init__(self, player.PLAYER)
        BaseControl.__init__(self)

        self.set_size_request(200, -1)
        self.set_margin_top(3)
        self.set_margin_bottom(3)

        if player.PLAYER.current is not None:
            self.on_playback_track_start(
                'playback_track_start', player.PLAYER, player.PLAYER.current
            )

            if player.PLAYER.is_paused():
                self.on_playback_toggle_pause(
                    'playback_toggle_pause', player.PLAYER, player.PLAYER.current
                )


control_types = [
    PreviousButtonControl,
    NextButtonControl,
    PlayPauseButtonControl,
    StopButtonControl,
    VolumeButtonControl,
    RestoreButtonControl,
    RatingControl,
    TrackSelectorControl,
    PlaylistButtonControl,
    ProgressButtonControl,
    ProgressBarControl,
]


def register():
    """
    Registers all control providers
    """
    for control_type in control_types:
        providers.register('minimode-controls', control_type)


def unregister():
    """
    Unregisters all control providers
    """
    for control_type in control_types:
        providers.unregister('minimode-controls', control_type)



# mono - Downmix Exaile's audio to one channel
# Copyright (C) 2013  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from gi.repository import Gst

import xl.providers
from xl.player.gst.gst_utils import ElementBin


class Mono(ElementBin):
    index = 90
    name = 'mono'

    def __init__(self):
        ElementBin.__init__(self, name=self.name)
        # self.elements[50] = Gst.ElementFactory.make('audioconvert', None)
        self.elements[60] = cf = Gst.ElementFactory.make('capsfilter', None)
        cf.props.caps = Gst.Caps.from_string('audio/x-raw,channels=1')
        self.setup_elements()


def enable(exaile):
    xl.providers.register('gst_audio_filter', Mono)


def disable(exaile):
    xl.providers.unregister('gst_audio_filter', Mono)


# vi: et sts=4 sw=4 tw=80



# Copyright (C) 2010 by Brian Parma
# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


from gi.repository import GLib
from gi.repository import Gtk

import time
import os
from xlgui import guiutil
from xl import event, player, settings, xdg, common
from functools import wraps
import logging
from . import macprefs
from .cellrenderers import CellRendererDays

import json

write = lambda x: json.dumps(x, indent=2)
read = json.loads

logger = logging.getLogger(__name__)

PLUGIN_ENABLED = False
TIMER_ID = None
MENU_ITEM = None
ALARM_CLOCK_MAIN = None
MY_BUILDER = None


def _init(prefsdialog, builder):
    '''Since I don't know if init() or enable() will be called first, save builder and setup UI if enabled() was already called.'''
    logger.debug('_init() called')
    global MY_BUILDER, ALARM_CLOCK_MAIN

    # note that we get a new builder everytime the prefs dialog is closed and re-opened
    MY_BUILDER = builder
    if ALARM_CLOCK_MAIN is not None:
        ALARM_CLOCK_MAIN.init_ui(builder)


def get_preferences_pane():
    '''Return prefs pane for preferences dialog'''
    logger.debug('get_preferences_pane() called')
    macprefs.init = _init
    return macprefs


def idle_add(f):
    '''Decorator that runs a function through GLib.idle_add'''

    @wraps(f)
    def idler(*args, **kwargs):
        GLib.idle_add(f, *args, **kwargs)

    return idler


###><><><### Alarm Clock Stuph ###><><><###


class Alarm:
    """
    Class for individual alarms.
    """

    def __init__(self, time="09:00", days=None, name="New Alarm", active=True, dict={}):
        self.active = active
        self.time = time
        self.name = name
        if days is None:
            self.days = [True] * 7
        else:
            self.days = days

        # For setting attributes by dictionary
        self.__dict__.update(dict)

    def on(self):
        self.active = True

    def off(self):
        self.active = False

    def toggle(self):
        if self.active:
            self.off()
        else:
            self.on()


class AlarmClock:
    """
    Class that handles the TreeView interaction and keeps track of alarms.
    """

    def __init__(self, exaile):
        self.RANG = {}  # Keep track of alarms that have gone off
        self.exaile = exaile
        #        self.view_window = None
        #        self.view = None

        # Create Model
        self.model = Gtk.ListStore(bool, str, str, object, str)

        # Load any saved alarms
        self.load_list()

    def _create_view(self):
        '''Create treeview to display model'''
        # Create View
        view = Gtk.TreeView()
        view.set_model(self.model)

        # setup view
        cr = Gtk.CellRendererToggle()
        cr.connect('toggled', self.enable_cb)
        col = Gtk.TreeViewColumn('Enabled', cr, active=0)
        view.append_column(col)

        cr = Gtk.CellRendererText()
        cr.connect('edited', self.text_edited, 1)
        cr.set_property('editable', True)
        col = Gtk.TreeViewColumn('Name', cr, text=1)
        view.append_column(col)

        cr = Gtk.CellRendererText()
        cr.connect('edited', self.text_edited, 2)
        cr.set_property('editable', True)
        col = Gtk.TreeViewColumn('Time', cr, text=2)
        view.append_column(col)

        # custom CellRenderer for Days Popup
        cr = CellRendererDays()
        cr.connect('days-changed', self.days_changed)
        cr.set_property('editable', True)
        col = Gtk.TreeViewColumn('Days', cr, days=3, text=4)
        view.append_column(col)

        return view

    def enable_cb(self, cell, path):
        '''Callback for toggling an alarm on/off'''
        active = self.model[path][0]
        self.model[path][0] = not active

        logger.debug(
            'Alarm {0} {1}abled.'.format(self.model[path][1], 'dis' if active else 'en')
        )

        # save change
        self.save_list()

    def init_ui(self, builder):
        '''Called by exaile to initialize prefs pane.  Set up pefs UI'''
        logger.debug('init_ui() called.')

        #        if self.view_window is not None:
        # already setup
        #            return

        # grab widgets
        view_window = builder.get_object('alarm_scrolledwindow')
        add_button = builder.get_object('add_button')
        del_button = builder.get_object('remove_button')

        # when a plugin is disabled and re-enabled, the preferences pane is not re-created until the prefs dialog is closed
        # so if we recycle our class and create a new one, we have to replace the old TreeView with our new one.
        #   NOTE: reloading the plugin from prefs page (DEBUG MODE) breaks this anyway
        child = view_window.get_child()
        view = self._create_view()
        if child is not None:
            logger.debug('stale treeview found, replacing...')
            guiutil.gtk_widget_replace(child, view)
        else:
            view_window.add(view)

        # signals
        add_button.connect('clicked', self.add_button)
        del_button.connect('clicked', self.delete_button, view.get_selection())

    def days_changed(self, cr, path, days):
        '''Callback for change of selected days for selected alarm'''
        # update model
        self.model[path][3] = days

        # update display
        days_str = ['Su', 'M', 'Tu', 'W', 'Th', 'F', 'Sa']
        self.model[path][4] = ','.join([days_str[i] for i in range(0, 7) if days[i]])

        # save changes
        self.save_list()

    def text_edited(self, cr, path, new_text, idx):
        '''Callback for edit of text columns (name and time)'''
        old_text = self.model[path][idx]
        if old_text == new_text:
            return  # No change

        if idx == 1:  # Name edit
            self.model[path][1] = new_text

            # save change
            self.save_list()

        elif idx == 2:  # Time edit
            # validate
            try:
                t = time.strptime(new_text, '%H:%M')
                new_text = time.strftime('%H:%M', t)  # ensure consistent 0-padding
            except ValueError:
                logger.warning('Invalid time format, use: HH:MM (24-hour)')
                return

            # update
            self.model[path][2] = new_text

            # save change
            self.save_list()

    def add_alarm(self, alarm):
        '''Add an alarm to the model'''
        # update display
        days_str = ['Su', 'M', 'Tu', 'W', 'Th', 'F', 'Sa']
        day_disp = ','.join([days_str[i] for i in range(0, 7) if alarm.days[i]])
        self.model.append([alarm.active, alarm.name, alarm.time, alarm.days, day_disp])

        # save list changes - NO, called by loader

    #        self.save_list()

    def add_button(self, widget):
        '''Callback for clicking add button'''
        # generate unique name
        names = [row[1] for row in self.model]
        base = 'Alarm'
        name = base
        i = 0
        while name in names:
            i = i + 1
            name = base + ' {0}'.format(i)

        # add the new alarm
        alarm = Alarm(name=name)
        self.add_alarm(alarm)

        # save list changes
        self.save_list()

    def delete_button(self, widget, selection):
        '''Callback for clicking the delete button'''
        # get selected row
        model, tree_iter = selection.get_selected()
        if tree_iter is not None:
            model.remove(tree_iter)

            # save list changes
            self.save_list()
        else:
            logger.info('No alarm selected for removal.')

    def load_list(self):
        '''Load alarms from file'''
        logger.debug('load_list() called.')
        path = os.path.join(xdg.get_data_dirs()[0], 'alarmlist.dat')
        try:
            # Load Alarm List from file.
            # Open in non-binary mode, because we are reading json
            # string.
            with open(path, 'r') as f:
                raw = f.read()
                try:
                    alist = _read(raw)
                    assert isinstance(
                        alist, list
                    )  # should be list of dicts (new format)

                except Exception:
                    try:
                        # try to import old format
                        for line in raw.strip().split('\n'):
                            a = Alarm(dict=eval(line, {'__builtin__': None}))
                            logger.debug(
                                'loaded alarm {0} ({1}) from file.'.format(
                                    a.name, a.time
                                )
                            )
                            self.add_alarm(a)

                        # force save in new format
                        logger.info('Old alarm file format found, converting.')
                        self.save_list()

                    except Exception as e:
                        logger.warning(
                            'Failed to load alarm data from file: {0}'.format(e)
                        )

                else:
                    for a in alist:
                        alarm = Alarm(dict=a)
                        logger.debug(
                            'loaded alarm {0} ({1}) from file.'.format(
                                alarm.name, alarm.time
                            )
                        )
                        self.add_alarm(alarm)

        except IOError as e:  # File might not exist
            logger.warning('Could not open file: {0}'.format(e.strerror))

    @idle_add
    def save_list(self):
        '''Save alarms to file'''
        logger.debug('save_list() called.')

        # Save List
        path = os.path.join(xdg.get_data_dirs()[0], 'alarmlist.dat')

        if len(self.model) > 0:
            alist = [
                {'active': row[0], 'name': row[1], 'time': row[2], 'days': row[3]}
                for row in self.model
            ]

            # Open in non-binary mode, because we are writing json
            # string.
            with open(path, 'w') as f:
                f.write(_write(alist))
                logger.debug('saving {0} alarms.'.format(len(alist)))


###><><><### Globals ###><><><###

# This is here because sometimes get_prefs_pane gets called before _enabled
# ALARM_CLOCK_MAIN = AlarmClock(exaile)


@common.threaded
def fade_in(main, exaile):
    '''Fade exaile's volume from min to max'''
    logger.debug('fade_in() called.')

    # pull settings
    temp_volume = (
        settings.get_option('plugin/multialarmclock/fade_min_volume', 0) / 100.0
    )
    fade_max_volume = (
        settings.get_option('plugin/multialarmclock/fade_max_volume', 100) / 100.0
    )
    fade_inc = settings.get_option('plugin/multialarmclock/fade_increment', 1) / 100.0
    time_per_inc = settings.get_option('plugin/multialarmclock/fade_time', 30) / (
        (fade_max_volume - temp_volume) / fade_inc
    )

    while temp_volume < fade_max_volume:
        logger.debug('set volume to {0}'.format(temp_volume))

        settings.set_option('player/volume', temp_volume)
        temp_volume += fade_inc
        time.sleep(time_per_inc)
        if player.PLAYER.is_paused() or not player.PLAYER.is_playing():
            return

    settings.set_option('player/volume', fade_max_volume)


def check_alarms(main, exaile):
    """
    Called every timeout.  If the plugin is not enabled, it does
    nothing.  If the current time matches the time specified and the
    current day is selected, it starts playing
    """
    if not main:
        return True  # TODO: new way?

    current = time.strftime("%H:%M", time.localtime())
    currentDay = int(time.strftime("%w", time.localtime()))

    # generate list of alarms from model
    alist = [
        Alarm(active=row[0], name=row[1], time=row[2], days=row[3])
        for row in main.model
    ]
    # print(current , [ a.time for a in alist if a.active ])
    for al in alist:
        if al.active and al.time == current and al.days[currentDay] is True:
            check = time.strftime("%m %d %Y %H:%M")  # clever...

            if check in main.RANG:
                logger.debug('Alarm {0} in RANG'.format(al.name))
                return True

            logger.info('Alarm {0} hit.'.format(al.name))

            # tracks to play?
            count = len(player.QUEUE)
            if player.QUEUE.current_playlist:
                count += len(player.QUEUE.current_playlist)
            else:
                count += len(exaile.gui.main.get_selected_page().playlist)

            if count == 0:
                logger.warning('No tracks queued for alarm to play.')
                return True

            if (
                player.PLAYER.is_playing()
            ):  # Check if there are songs in playlist and if it is already playing
                logger.info('Alarm hit, but already playing')
                return True

            if settings.get_option('plugin/multialarmclock/fading_on'):
                fade_in(main, exaile)

            if settings.get_option('plugin/multialarmclock/restart_playlist_on'):
                logger.debug('try to restart playlist')
                if player.QUEUE.current_playlist:
                    player.QUEUE.current_playlist.set_current_position(-1)
                else:
                    player.QUEUE.set_current_playlist(
                        exaile.gui.main.get_selected_page().playlist
                    )

            player.QUEUE.play()

            main.RANG[check] = True

    return True


###><><><### Plugin Handling Functions ###><><><###


def enable(exaile):
    '''Called by exaile to enable plugin'''
    if exaile.loading:
        logger.debug('waitin for loaded event')
        event.add_callback(_enable, 'exaile_loaded')
    else:
        _enable(None, exaile, None)


@idle_add
def _enable(stuff, exaile, junk):
    """
    Enable plugin.  Start timer and create class.
    """
    logger.debug('_enable called')
    global TIMER_ID, MENU_ITEM, ALARM_CLOCK_MAIN, MY_BUILDER

    if ALARM_CLOCK_MAIN is None:
        ALARM_CLOCK_MAIN = AlarmClock(exaile)

    main = ALARM_CLOCK_MAIN

    if MY_BUILDER is not None:
        # '''Since I don't know if init() or enable() will be called first, save builder and setup UI if enabled() was already called.'''
        main.init_ui(MY_BUILDER)

    TIMER_ID = GLib.timeout_add_seconds(5, check_alarms, main, exaile)


def disable(exaile):
    """
    Called when plugin is unloaded.  Stop timer.
    """
    global TIMER_ID, MENU_ITEM, ALARM_CLOCK_MAIN, MY_BUILDER

    # Cleanup
    if TIMER_ID is not None:
        GLib.source_remove(TIMER_ID)
        TIMER_ID = None

    if ALARM_CLOCK_MAIN is not None:
        ALARM_CLOCK_MAIN = None


#   disable/enable doesn't re-call init(), so if we scrap and re-create the class, we wont' get our Gtk.Builder back, and there will
#   be a disconnect between the Alarm class and the UI in the prefs page
#   NOTE: reloading the plugin from prefs page (DEBUG MODE) breaks this anyway
#    if MY_BUILDER is not None:
#        MY_BUILDER = None



# -*- coding: utf-8

from mox3 import mox

from xl.trax import search
from xl.trax import track
import pytest


def test_search_result_track_get_track():
    val = 'foo'
    search_result_track = search.SearchResultTrack(val)
    assert search_result_track.track == val, search_result_track.track


def get_search_result_track():
    tr = track.Track('file:///foo')
    return search.SearchResultTrack(tr)


class TestMatcher:
    def setup(self):
        self.mox = mox.Mox()
        self.strack = get_search_result_track()
        self.strack.track.set_tag_raw('artist', ['foo', 'bar'])

    def teardown(self):
        self.mox.UnsetStubs()

    def test_match_list_true(self):
        self.mox.StubOutWithMock(search._Matcher, '_matches')
        search._Matcher._matches(mox.IsA(str)).AndReturn(True)
        self.mox.ReplayAll()
        matcher = search._Matcher('artist', 'bar', lambda x: x)
        assert matcher.match(self.strack)
        self.mox.VerifyAll()

    def test_match_list_false(self):
        self.mox.StubOutWithMock(search._Matcher, '_matches')
        # ensure that both tags are checked
        search._Matcher._matches(mox.IsA(str)).AndReturn(False)
        search._Matcher._matches(mox.IsA(str)).AndReturn(False)
        self.mox.ReplayAll()
        matcher = search._Matcher('artist', 'bar', lambda x: x)
        assert not matcher.match(self.strack)
        self.mox.VerifyAll()

    def test_match_list_none(self):
        self.mox.StubOutWithMock(search._Matcher, '_matches')
        search._Matcher._matches(None).AndReturn(True)
        self.mox.ReplayAll()
        matcher = search._Matcher('album', None, lambda x: x)
        assert matcher.match(self.strack)
        self.mox.VerifyAll()

    def test_matches(self):
        matcher = search._Matcher('album', None, lambda x: x)
        with pytest.raises(NotImplementedError):
            matcher._matches('foo')


class TestExactMatcher:
    def setup(self):
        self.str = get_search_result_track()

    def test_exact_matcher_true(self):
        matcher = search._ExactMatcher('album', 'Foo', lambda x: x)
        self.str.track.set_tag_raw('album', 'Foo')
        assert matcher.match(self.str)

    def test_exact_matcher_false(self):
        matcher = search._ExactMatcher('album', 'Foo', lambda x: x)
        self.str.track.set_tag_raw('album', 'FoO')
        assert not matcher.match(self.str)


class TestInMatcher:
    def setup(self):
        self.str = get_search_result_track()

    def test_in_matcher_none(self):
        matcher = search._InMatcher('album', 'Foo', lambda x: x)
        self.str.track.set_tag_raw('album', None)
        assert not matcher.match(self.str)

    def test_in_matcher_true(self):
        matcher = search._InMatcher('album', 'hello', lambda x: x)
        self.str.track.set_tag_raw('album', 'Foohelloworld')
        assert matcher.match(self.str)

    def test_in_matcher_error(self):
        matcher = search._InMatcher('album', 2, lambda x: x)
        self.str.track.set_tag_raw('album', 'Foohelloworld')
        assert not matcher.match(self.str)

    def test_in_matcher_false(self):
        matcher = search._InMatcher('album', 'hello', lambda x: x)
        self.str.track.set_tag_raw('album', 'Fooheloworld')
        assert not matcher.match(self.str)


class TestGtLtMatchers:
    def setup(self):
        self.str = get_search_result_track()

    def test_gt_bitrate_matcher_true(self):
        matcher = search._GtMatcher('__bitrate', 100000, lambda x: x)
        self.str.track.set_tag_raw('__bitrate', 128000)
        assert matcher.match(self.str)

    def test_gt_bitrate_matcher_false(self):
        matcher = search._GtMatcher('__bitrate', 100000, lambda x: x)
        self.str.track.set_tag_raw('__bitrate', 28000)
        assert not matcher.match(self.str)

    def test_lt_bitrate_matcher_true(self):
        matcher = search._LtMatcher('__bitrate', 100000, lambda x: x)
        self.str.track.set_tag_raw('__bitrate', 28000)
        assert matcher.match(self.str)

    def test_lt_bitrate_matcher_false(self):
        matcher = search._LtMatcher('__bitrate', 100000, lambda x: x)
        self.str.track.set_tag_raw('__bitrate', 128000)
        assert not matcher.match(self.str)


class TestMetaMatcherClasses:
    class _Matcher:
        def __init__(self, val):
            self.val = val

        def matches(self, val):
            return self.val

        def match(self, val):
            return self.val


class TestNotMetaMatcher(TestMetaMatcherClasses):
    def test_true(self):
        matcher = self._Matcher(True)
        matcher = search._NotMetaMatcher(matcher)
        assert not matcher.match('foo')

    def test_false(self):
        matcher = self._Matcher(False)
        matcher = search._NotMetaMatcher(matcher)
        assert matcher.match('foo')


class TestOrMetaMatcher(TestMetaMatcherClasses):
    def test_true_true(self):
        matcher_1 = self._Matcher(True)
        matcher_2 = self._Matcher(True)
        matcher = search._OrMetaMatcher(matcher_1, matcher_2)
        assert matcher.match('foo')

    def test_true_false(self):
        matcher_1 = self._Matcher(True)
        matcher_2 = self._Matcher(False)
        matcher = search._OrMetaMatcher(matcher_1, matcher_2)
        assert matcher.match('foo')

    def test_false_true(self):
        matcher_1 = self._Matcher(False)
        matcher_2 = self._Matcher(True)
        matcher = search._OrMetaMatcher(matcher_1, matcher_2)
        assert matcher.match('foo')

    def test_false_false(self):
        matcher_1 = self._Matcher(False)
        matcher_2 = self._Matcher(False)
        matcher = search._OrMetaMatcher(matcher_1, matcher_2)
        assert not matcher.match('foo')


class TestMultiMetaMatcher(TestMetaMatcherClasses):
    def test_true(self):
        matcher = [self._Matcher(True)] * 10
        matcher = search._MultiMetaMatcher(matcher)
        assert matcher.match('foo')

    def test_false(self):
        matcher = [self._Matcher(True)] * 10 + [self._Matcher(False)]
        matcher = search._MultiMetaMatcher(matcher)
        assert not matcher.match('foo')


class TestManyMultiMetaMatcher(TestMetaMatcherClasses):
    def test_true(self):
        matcher = [self._Matcher(True)] * 10 + [self._Matcher(False)]
        for match in matcher:
            match.tag = 'artist'
        matcher = search._ManyMultiMetaMatcher(matcher)
        assert matcher.match('foo')

    def test_false(self):
        matcher = [self._Matcher(False)] * 10
        for match in matcher:
            match.tag = 'artist'
        matcher = search._ManyMultiMetaMatcher(matcher)
        assert not matcher.match('foo')


class TestTracksMatcher:
    def setup(self):
        self.str = get_search_result_track()

    def test_in_matcher(self):
        matcher = search.TracksMatcher("artist=foo")
        assert len(matcher.matchers) == 1
        match = matcher.matchers[0]
        self.match_is_type(match, search._InMatcher)
        assert match.tag == 'artist'
        assert match.content == 'foo'

    def test_exact_matcher(self):
        matcher = search.TracksMatcher("artist==foo")
        assert len(matcher.matchers) == 1
        match = matcher.matchers[0]
        self.match_is_type(match, search._ExactMatcher)
        assert match.tag == 'artist'
        assert match.content == 'foo'

    def match_is_type(self, match, expected):
        assert isinstance(match, expected), match

    def test_not_matcher(self):
        matcher = search.TracksMatcher("! foo", keyword_tags=['artist'])
        match = matcher
        # NotMetaMatcher
        assert len(match.matchers) == 1
        match = matcher.matchers[0]
        self.match_is_type(match, search._NotMetaMatcher)
        # MultiMetaMatcher
        match = match.matcher
        self.match_is_type(match, search._MultiMetaMatcher)
        # ManyMultiMetaMatcher
        assert len(match.matchers) == 1
        match = match.matchers[0]
        self.match_is_type(match, search._ManyMultiMetaMatcher)
        # InMatcher
        assert len(match.matchers) == 1
        match = match.matchers[0]
        self.match_is_type(match, search._InMatcher)
        assert match.tag == 'artist'
        assert match.content == 'foo'

    def test_or_matcher(self):
        matcher = search.TracksMatcher("foo | bar", keyword_tags=['artist'])
        match = matcher
        # OrMetaMatcher
        assert len(match.matchers) == 1
        match = matcher.matchers[0]
        self.match_is_type(match, search._OrMetaMatcher)
        # MultiMetaMatcher
        assert match.left
        assert match.right
        self.match_is_type(match.left, search._MultiMetaMatcher)
        self.match_is_type(match.right, search._MultiMetaMatcher)
        # ManyMultiMetaMatcher
        assert len(match.left.matchers) == 1
        assert len(match.right.matchers) == 1
        match_left = match.left.matchers[0]
        match_right = match.right.matchers[0]
        self.match_is_type(match_left, search._ManyMultiMetaMatcher)
        self.match_is_type(match_right, search._ManyMultiMetaMatcher)
        # InMatcher
        assert len(match_left.matchers) == 1
        assert len(match_right.matchers) == 1
        match_left = match_left.matchers[0]
        match_right = match_right.matchers[0]
        self.match_is_type(match_left, search._InMatcher)
        self.match_is_type(match_right, search._InMatcher)
        assert match_left.tag == 'artist'
        assert match_right.tag == 'artist'
        if match_left.content == 'foo':
            if match_right.content != 'bar':
                assert not "We lost a search term on an or"
        elif match_left.content == 'bar':
            if match_right.content != 'foo':
                assert not "We lost a search term on an or"
        else:
            assert not "We lost both parts of an or"

    def test_paren_matcher(self):
        matcher = search.TracksMatcher("( foo | bar )", keyword_tags=['artist'])
        match = matcher
        # MultiMetaMatcher
        assert len(match.matchers) == 1
        match = matcher.matchers[0]
        self.match_is_type(match, search._MultiMetaMatcher)

        assert len(match.matchers) == 1
        match = match.matchers[0]
        self.match_is_type(match, search._OrMetaMatcher)

        # This is the same code as the OrMetaMatcher
        assert match.left
        assert match.right
        self.match_is_type(match.left, search._MultiMetaMatcher)
        self.match_is_type(match.right, search._MultiMetaMatcher)
        # ManyMultiMetaMatcher
        assert len(match.left.matchers) == 1
        assert len(match.right.matchers) == 1
        match_left = match.left.matchers[0]
        match_right = match.right.matchers[0]
        self.match_is_type(match_left, search._ManyMultiMetaMatcher)
        self.match_is_type(match_right, search._ManyMultiMetaMatcher)
        # InMatcher
        assert len(match_left.matchers) == 1
        assert len(match_right.matchers) == 1
        match_left = match_left.matchers[0]
        match_right = match_right.matchers[0]
        self.match_is_type(match_left, search._InMatcher)
        self.match_is_type(match_right, search._InMatcher)
        assert match_left.tag == 'artist'
        assert match_right.tag == 'artist'
        if match_left.content == 'foo':
            if match_right.content != 'bar':
                assert not "We lost a search term on an or"
        elif match_left.content == 'bar':
            if match_right.content != 'foo':
                assert not "We lost a search term on an or"
        else:
            assert not "We lost both parts of an or"

    def test_match_true(self):
        matcher = search.TracksMatcher("foo", keyword_tags=['artist'])
        self.str.track.set_tag_raw('artist', 'foo')
        assert matcher.match(self.str)
        assert self.str.on_tags == ['artist']

    def test_match_true_tag(self):
        matcher = search.TracksMatcher("artist=foo")
        self.str.track.set_tag_raw('artist', 'foo')
        assert matcher.match(self.str)
        assert self.str.on_tags == ['artist']

    def test_match_true_case_insensitive(self):
        matcher = search.TracksMatcher("artist=FoO", case_sensitive=False)
        self.str.track.set_tag_raw('artist', 'foo')
        assert matcher.match(self.str)
        assert self.str.on_tags == ['artist']

    def test_match_true_none(self):
        matcher = search.TracksMatcher("artist==__null__")
        self.str.track.set_tag_raw('artist', '')
        assert matcher.match(self.str)
        assert self.str.on_tags == ['artist']

    def test_match_false(self):
        matcher = search.TracksMatcher("foo", keyword_tags=['artist'])
        self.str.track.set_tag_raw('artist', 'bar')
        assert not matcher.match(self.str)


class TestSearchTracks:
    def test_search_tracks(self):
        matcher = search.TracksMatcher("foo", keyword_tags=['artist'])
        tracks = [track.Track(x) for x in ('foo', 'bar', 'baz', 'quux')]
        tracks = [search.SearchResultTrack(tr) for tr in tracks]
        tracks[0].track.set_tag_raw('artist', 'foooo')
        tracks[2].track.set_tag_raw('artist', 'foooooo')
        gen = search.search_tracks(tracks, [matcher])
        assert next(gen) == tracks[0]
        assert next(gen) == tracks[2]
        with pytest.raises(StopIteration):
            next(gen)

    def test_take_not_srt(self):
        matcher = search.TracksMatcher("foo", keyword_tags=['artist'])
        tracks = [track.Track(x) for x in ('foo', 'bar', 'baz', 'quux')]
        tracks[0].set_tag_raw('artist', 'foooo')
        tracks[2].set_tag_raw('artist', 'foooooo')
        gen = search.search_tracks(tracks, [matcher])
        assert next(gen).track == tracks[0]
        assert next(gen).track == tracks[2]
        with pytest.raises(StopIteration):
            next(gen)

    def test_search_tracks_from_string(self):
        tracks = [track.Track(x) for x in ('foo', 'bar', 'baz', 'quux')]
        tracks[0].set_tag_raw('artist', 'foooo')
        tracks[2].set_tag_raw('artist', 'foooooo')
        gen = search.search_tracks_from_string(tracks, 'foo', keyword_tags=['artist'])
        assert next(gen).track == tracks[0]
        assert next(gen).track == tracks[2]
        with pytest.raises(StopIteration):
            next(gen)

    @pytest.mark.parametrize("sstr", ["motley crue", "mötley crüe", "motley crüe"])
    def test_search_tracks_ignore_diacritic_from_string(self, sstr):
        """Ensure that searching for tracks with diacritics return
        appropriately normalized results"""
        tracks = [track.Track(x) for x in ('foo', 'bar', 'baz', 'quux')]
        tracks[0].set_tag_raw('artist', 'motley crue')
        tracks[1].set_tag_raw('artist', 'rubbish')
        tracks[2].set_tag_raw('artist', 'motley crüe')

        gen = search.search_tracks_from_string(tracks, sstr, keyword_tags=['artist'])

        assert next(gen).track == tracks[0]
        assert next(gen).track == tracks[2]
        with pytest.raises(StopIteration):
            next(gen)

    def test_search_tracks_with_unicodemark_from_string(self):
        tracks = [track.Track(x) for x in ('foo', 'bar', 'baz', 'quux')]
        tracks[0].set_tag_raw('artist', 'foooo')
        tracks[2].set_tag_raw('artist', '中')

        # the weird character is normalized, so you can't search based on that
        gen = search.search_tracks_from_string(tracks, '中', keyword_tags=['artist'])

        assert next(gen).track == tracks[2]
        with pytest.raises(StopIteration):
            next(gen)

    def test_search_tracks_with_int_from_string(self):
        # unlike mp3, mp4 will return integers for BPM.. make sure that works
        tracks = [track.Track(x) for x in ('foo', 'bar', 'baz', 'quux')]
        tracks[1].set_tag_raw('bpm', '2')
        tracks[2].set_tag_raw('bpm', 2)
        gen = search.search_tracks_from_string(tracks, '2', keyword_tags=['bpm'])
        assert next(gen).track == tracks[1]
        assert next(gen).track == tracks[2]
        with pytest.raises(StopIteration):
            next(gen)



import os


def test_mp3_exists(test_tracks):
    assert test_tracks.get('.mp3')


def test_all_tracks(test_track):
    assert os.path.exists(test_track.filename), (
        "%s does not exist" % test_track.filename
    )

    assert test_track.uri.startswith('file:///')


def test_writable_tracks(writeable_track):
    assert os.path.exists(writeable_track.filename), (
        "%s does not exist" % writeable_track.filename
    )

    assert writeable_track.uri.startswith('file:///')



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import xl.unicode
from xl.metadata._base import CaseInsensitveBaseFormat, CoverImage
from mutagen import oggvorbis, oggopus
from mutagen.flac import Picture
import base64


class OggFormat(CaseInsensitveBaseFormat):
    MutagenType = oggvorbis.OggVorbis
    tag_mapping = {
        'bpm': 'tempo',
        'comment': 'description',
        'cover': 'metadata_block_picture',
    }
    writable = True

    def _get_tag(self, raw, tag):
        value = CaseInsensitveBaseFormat._get_tag(self, raw, tag)
        if value and tag == 'metadata_block_picture':
            new_value = []
            for v in value:
                picture = Picture(base64.b64decode(v))
                new_value.append(
                    CoverImage(
                        type=picture.type,
                        desc=picture.desc,
                        mime=picture.mime,
                        data=picture.data,
                    )
                )
            value = new_value
        return value

    def _set_tag(self, raw, tag, value):
        if tag == 'metadata_block_picture':
            new_value = []
            for v in value:
                picture = Picture()
                picture.type = v.type
                picture.desc = v.desc
                picture.mime = v.mime
                picture.data = v.data
                tmp = base64.b64encode(picture.write())
                tmp = tmp.decode('ascii')  # needs to be a str
                new_value.append(tmp)
            value = new_value
        else:
            # vorbis has text based attributes, so convert everything to unicode
            value = [xl.unicode.to_unicode(v) for v in value]
        CaseInsensitveBaseFormat._set_tag(self, raw, tag, value)


class OggOpusFormat(OggFormat):
    MutagenType = oggopus.OggOpus



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from xl import playlist, event, providers, trax


class RadioManager(providers.ProviderHandler):
    """
    Radio Station Manager

    Simple usage:

    >>> manager = RadioManager()
    >>> manager.add_station(RadioStation())
    >>> lists = manager.get_lists('test_station')
    >>> pl = lists[0].get_playlist()
    >>> print(pl.get_tracks()[0]['title'][0])
    Test Track
    >>>
    """

    def __init__(self):
        """
        Initializes the radio manager
        """
        providers.ProviderHandler.__init__(self, "radio")
        self.stations = {}

    def add_station(self, station):
        """
        Adds a station to the manager

        @param station: The station to add
        """
        providers.register(self.servicename, station)

    def on_provider_added(self, station):
        if station.name not in self.stations:
            self.stations[station.name] = station
            event.log_event('station_added', self, station)

    def remove_station(self, station):
        """
        Removes a station from the manager

        @param station: The station to remvoe
        """
        providers.unregister(self.servicename, station)

    def on_provider_removed(self, station):
        if station.name in self.stations:
            del self.stations[station.name]
            event.log_event('station_removed', self, station)

    def search(self, station, keyword):
        if station in self.stations:
            return self.stations[station].search(keyword)
        else:
            return None

    def get_lists(self, station, no_cache=False):
        """
        Loads the lists for the specified station

        @param station: The name of the station
        """
        if station in self.stations:
            return self.stations[station].get_lists(no_cache=no_cache)
        else:
            return None

    def load_lists(self, station):
        """
        Gets the rlists for the specified station

        @param station:  The name of the station
        """
        if station in self.stations:
            return self.stations[station].load_lists()
        else:
            return None


class RadioList:
    def __init__(self, name, station=None):
        """
        Initializes the rlist
        """
        self.name = name
        self.station = station

    def set_name(self, name):
        self.name = name

    def get_name(self):
        return self.name

    def get_items(self, no_cache=False):
        """
        Returns subrlists
        """
        return []

    def __str__(self):
        """
        Returns a string representation of the list
        """
        return self.name


class RadioItem:
    """
    Radio Items
    """

    def __init__(self, name, station=None):
        """
        Initializes the radio item
        """
        self.name = name
        self.station = station

    def get_playlist(self):
        tr = trax.Track()
        tr['title'] = 'Test Track'
        pl = playlist.Playlist('Test Playlist')
        pl.add_tracks([tr])
        return pl

    def __str__(self):
        """
        Returns a string representation of the item
        """
        return self.name


class RadioStation:
    name = 'test_station'

    def __init__(self):
        """
        Initialize the radio station
        """
        pass

    def get_lists(self, no_cache=False):
        """
        Returns the rlists for this radio station
        """
        return [RadioItem('TestCategory')]

    def search(self, keyword):
        return None

    def __str__(self):
        """
        Returns a stream representation of this station
        """
        name = self.__class__.__name__
        name = name.replace('RadioStation', ' Radio')
        return name



import spydaap
import spydaap.parser.mp3
import spydaap.parser.ogg
import spydaap.parser.flac

spydaap.parsers = [
    spydaap.parser.mp3.Mp3Parser(),
    spydaap.parser.flac.FlacParser(),
    spydaap.parser.ogg.OggParser(),
]

# to process .mov files
# from spydaap.parser import mp3,mov

# spydaap.server_name = "spydaap"
# spydaap.port = 3689

# top path to scan for media
# spydaap.media_path = "media"

# spydaap.cache_dir = 'cache'

spydaap.container_list.append(
    spydaap.playlists.Genre('reggae', ["reggae", "reggae: dub", "dub"])
)
spydaap.container_list.append(
    spydaap.playlists.Recent('last 30 days', (60 * 60 * 24 * 30))
)
spydaap.container_list.append(spydaap.playlists.YearRange('1970s', 1970, 1979))
spydaap.container_list.append(spydaap.playlists.Rating('top rated', 100))



# Copyright (C) 2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gdk
from gi.repository import Gtk

import re
from datetime import datetime
from typing import List

from xl.nls import gettext as _
from xl import event, providers, settings
from xl.playlist import Playlist, PlaylistManager
from xlgui.widgets import menu
from xlgui.accelerators import Accelerator
from xlgui.widgets.notebook import (
    SmartNotebook,
    NotebookTab,
    NotebookAction,
    NotebookActionService,
)
from xlgui.widgets.playlist import PlaylistPage
from xlgui.widgets.queue import QueuePage

import logging

logger = logging.getLogger(__name__)


class NewPlaylistNotebookAction(NotebookAction, Gtk.Button):
    """
    Playlist notebook action which allows for creating new playlists
    regularly as well as by dropping tracks, files and directories on it
    """

    __gsignals__ = {'clicked': 'override'}
    name = 'new-playlist'
    position = Gtk.PackType.START

    def __init__(self, notebook):
        NotebookAction.__init__(self, notebook)
        Gtk.Button.__init__(self)

        self.set_image(Gtk.Image.new_from_icon_name('tab-new', Gtk.IconSize.BUTTON))
        self.set_relief(Gtk.ReliefStyle.NONE)

        self.__default_tooltip_text = _('New Playlist')
        self.__drag_tooltip_text = _('Drop here to create a new playlist')
        self.set_tooltip_text(self.__default_tooltip_text)

        self.drag_dest_set(Gtk.DestDefaults.ALL, [], Gdk.DragAction.COPY)
        self.drag_dest_add_uri_targets()

        self.connect('drag-motion', self.on_drag_motion)
        self.connect('drag-leave', self.on_drag_leave)
        self.connect('drag-data-received', self.on_drag_data_received)

    def do_clicked(self):
        """
        Triggers creation of a new playlist
        """
        self.notebook.create_new_playlist()

    def on_drag_motion(self, widget, context, x, y, time):
        """
        Updates the tooltip during drag operations
        """
        self.set_tooltip_text(self.__drag_tooltip_text)

    def on_drag_leave(self, widget, context, time):
        """
        Restores the original tooltip
        """
        self.set_tooltip_text(self.__default_tooltip_text)

    def on_drag_data_received(self, widget, context, x, y, selection, info, time):
        """
        Handles dropped data
        """
        tab = self.notebook.create_new_playlist()
        # Forward signal to the PlaylistView in the newly added tab
        tab.page.view.emit('drag-data-received', context, x, y, selection, info, time)


providers.register('playlist-notebook-actions', NewPlaylistNotebookAction)


class PlaylistNotebook(SmartNotebook):
    def __init__(self, manager_name, player, hotkey):
        SmartNotebook.__init__(self)

        self.tab_manager = PlaylistManager(manager_name)
        self.manager_name = manager_name
        self.player = player

        # For saving closed tab history
        self._moving_tab = False
        self.tab_history = []
        self.history_counter = 90000  # to get unique (reverse-ordered) item names

        # Build static menu entries
        item = menu.simple_separator('clear-sep', [])
        item.register('playlist-closed-tab-menu', self)

        item = menu.simple_menu_item(
            'clear-history',
            ['clear-sep'],
            _("_Clear Tab History"),
            'edit-clear-all',
            self.clear_closed_tabs,
        )
        item.register('playlist-closed-tab-menu', self)

        # Simple factory for 'Recently Closed Tabs' MenuItem
        submenu = menu.ProviderMenu('playlist-closed-tab-menu', self)

        def factory(menu_, parent, context):
            if self.page_num(parent) == -1:
                return None
            item = Gtk.MenuItem.new_with_mnemonic(_("Recently Closed _Tabs"))
            if len(self.tab_history) > 0:
                item.set_submenu(submenu)
            else:
                item.set_sensitive(False)
            return item

        # Add menu to tab context menu
        item = menu.MenuItem('%s-tab-history' % manager_name, factory, ['tab-close'])
        item.register('playlist-tab-context-menu')

        # Add menu to View menu
        # item = menu.MenuItem('tab-history', factory, ['clear-playlist'])
        # providers.register('menubar-view-menu', item)

        # setup notebook actions
        self.actions = NotebookActionService(self, 'playlist-notebook-actions')

        # Add hotkey
        self.accelerator = Accelerator(
            hotkey, _('Restore closed tab'), lambda *x: self.restore_closed_tab(0)
        )
        providers.register('mainwindow-accelerators', self.accelerator)

        # Load saved tabs
        self.load_saved_tabs()

        self.tab_placement_map = {
            'left': Gtk.PositionType.LEFT,
            'right': Gtk.PositionType.RIGHT,
            'top': Gtk.PositionType.TOP,
            'bottom': Gtk.PositionType.BOTTOM,
        }

        self.connect('page-added', self.on_page_added)
        self.connect('page-removed', self.on_page_removed)

        self.on_option_set('gui_option_set', settings, 'gui/show_tabbar')
        self.on_option_set('gui_option_set', settings, 'gui/tab_placement')
        event.add_ui_callback(self.on_option_set, 'gui_option_set')

    def create_tab_from_playlist(self, playlist):
        """
        Create a tab that will contain the passed-in playlist

        :param playlist: The playlist to create tab from
        :type playlist: :class:`xl.playlist.Playlist`
        """
        page = PlaylistPage(playlist, self.player)
        tab = NotebookTab(self, page)
        self.add_tab(tab, page)
        return tab

    def create_new_playlist(self):
        """
        Create a new tab containing a blank playlist.
        The tab will be automatically given a unique name.
        """
        seen = []
        default_playlist_name = _('Playlist %d')
        # Split into 'Playlist ' and ''
        default_name_parts = default_playlist_name.split('%d')

        for n in range(self.get_n_pages()):
            page = self.get_nth_page(n)
            name = page.get_page_name()
            name_parts = [
                # 'Playlist 99' => 'Playlist '
                name[0 : len(default_name_parts[0])],
                # 'Playlist 99' => ''
                name[len(name) - len(default_name_parts[1]) :],
            ]

            # Playlist name matches our format
            if name_parts == default_name_parts:
                # Extract possible number between name parts
                number = name[len(name_parts[0]) : len(name) - len(name_parts[1])]

                try:
                    number = int(number)
                except ValueError:
                    pass
                else:
                    seen += [number]

        seen.sort()
        n = 1

        while True:
            if n not in seen:
                break
            n += 1

        playlist = Playlist(default_playlist_name % n)

        return self.create_tab_from_playlist(playlist)

    def add_default_tab(self):
        return self.create_new_playlist()

    def load_saved_tabs(self):
        names = self.tab_manager.list_playlists()
        if not names:
            return

        count = -1
        count2 = 0
        names.sort()
        # holds the order#'s of the already added tabs
        added_tabs = {}
        name_re = re.compile(r'^order(?P<tab>\d+)\.(?P<tag>[^.]*)\.(?P<name>.*)$')
        for i, name in enumerate(names):
            match = name_re.match(name)
            if not match or not match.group('tab') or not match.group('name'):
                logger.error("`%r` did not match valid playlist file", name)
                continue

            logger.debug("Adding playlist %d: %s", i, name)
            logger.debug(
                "Tab:%s; Tag:%s; Name:%s",
                match.group('tab'),
                match.group('tag'),
                match.group('name'),
            )
            pl = self.tab_manager.get_playlist(name)
            pl.name = match.group('name')

            if match.group('tab') not in added_tabs:
                self.create_tab_from_playlist(pl)
                added_tabs[match.group('tab')] = pl
            pl = added_tabs[match.group('tab')]

            if match.group('tag') == 'current':
                count = i
                if self.player.queue.current_playlist is None:
                    self.player.queue.set_current_playlist(pl)
            elif match.group('tag') == 'playing':
                count2 = i
                self.player.queue.set_current_playlist(pl)

        # If there's no selected playlist saved, use the currently
        # playing
        if count == -1:
            count = count2

        self.set_current_page(count)

    def save_current_tabs(self):
        """
        Saves the open tabs
        """
        # first, delete the current tabs
        names = self.tab_manager.list_playlists()
        for name in names:
            logger.debug("Removing tab %s", name)
            self.tab_manager.remove_playlist(name)

        # TODO: make this generic enough to save other kinds of tabs
        for n, page in enumerate(self):
            if not isinstance(page, PlaylistPage):
                continue

            tag = ''

            if page.playlist is self.player.queue.current_playlist:
                tag = 'playing'
            elif n == self.get_current_page():
                tag = 'current'

            page.playlist.name = 'order%d.%s.%s' % (n, tag, page.playlist.name)
            logger.debug('Saving tab %r', page.playlist.name)

            try:
                self.tab_manager.save_playlist(page.playlist, True)
            except Exception:
                # an exception here could cause exaile to be unable to quit.
                # Catch all exceptions.
                logger.exception("Error saving tab %r", page.playlist.name)

    def show_current_track(self):
        """
        Tries to find the currently playing track
        and selects it and its containing tab page
        """
        for n, page in enumerate(self):
            if not isinstance(page, PlaylistPage):
                continue

            if page.playlist is not self.player.queue.current_playlist:
                continue

            self.set_current_page(n)
            page.view.scroll_to_cell(page.playlist.current_position)
            page.view.set_cursor(page.playlist.current_position)
            return True

    def on_page_added(self, notebook, child, page_number):
        """
        Updates appearance on page add
        """
        if self.get_n_pages() > 1:
            # Enforce tabbar visibility
            self.set_show_tabs(True)

    def on_page_removed(self, notebook, child, page_number):
        """
        Updates appearance on page removal
        """
        if self.get_n_pages() == 1:
            self.set_show_tabs(settings.get_option('gui/show_tabbar', True))

        # closed tab history
        if not self._moving_tab:

            if settings.get_option('gui/save_closed_tabs', True) and isinstance(
                child, PlaylistPage
            ):
                self.save_closed_tab(child.playlist)

            # Destroy it unless it's the queue page
            if not isinstance(child, QueuePage):
                child.destroy()

    def restore_closed_tab(self, pos=None, playlist=None, item_name=None):
        ret = self.remove_closed_tab(pos, playlist, item_name)
        if ret is not None:
            self.create_tab_from_playlist(ret[0])

    def save_closed_tab(self, playlist):
        # don't let the list grow indefinitely
        if len(self.tab_history) > settings.get_option('gui/max_closed_tabs', 10):
            self.remove_closed_tab(-1)  # remove last item

        item_name = 'playlist%05d' % self.history_counter
        close_time = datetime.now()
        # define a MenuItem factory that supports dynamic labels

        def factory(menu_, parent, context):
            item = None

            dt = datetime.now() - close_time
            if dt.seconds > 60:
                display_name = _(
                    '{playlist_name} ({track_count} tracks, closed {minutes} min ago)'
                ).format(
                    playlist_name=playlist.name,
                    track_count=len(playlist),
                    minutes=dt.seconds // 60,
                )
            else:
                display_name = _(
                    '{playlist_name} ({track_count} tracks, closed {seconds} sec ago)'
                ).format(
                    playlist_name=playlist.name,
                    track_count=len(playlist),
                    seconds=dt.seconds,
                )
            item = Gtk.ImageMenuItem.new_with_mnemonic(display_name)
            item.set_image(
                Gtk.Image.new_from_icon_name('music-library', Gtk.IconSize.MENU)
            )

            # Add accelerator to top item
            if self.tab_history[0][1].name == item_name:
                key, mods = Gtk.accelerator_parse(self.accelerator.keys)
                item.add_accelerator(
                    'activate', menu.FAKEACCELGROUP, key, mods, Gtk.AccelFlags.VISIBLE
                )

            item.connect(
                'activate', lambda w: self.restore_closed_tab(item_name=item_name)
            )

            return item

        # create menuitem
        item = menu.MenuItem(item_name, factory, [])
        providers.register('playlist-closed-tab-menu', item, self)
        self.history_counter -= 1

        # add
        self.tab_history.insert(0, (playlist, item))

    def get_closed_tab(self, pos=None, playlist=None, item_name=None):
        if pos is not None:
            try:
                return self.tab_history[pos]
            except IndexError:
                return None
        elif playlist is not None:
            for (pl, item) in self.tab_history:
                if pl == playlist:
                    return (pl, item)
        elif item_name is not None:
            for (pl, item) in self.tab_history:
                if item.name == item_name:
                    return (pl, item)

        return None
        # remove from menus

    def remove_closed_tab(self, pos=None, playlist=None, item_name=None):
        ret = self.get_closed_tab(pos, playlist, item_name)
        if ret is not None:
            self.tab_history.remove(ret)
            providers.unregister('playlist-closed-tab-menu', ret[1], self)
        return ret

    def clear_closed_tabs(self, widget, name, parent, context):
        for i in range(len(self.tab_history)):
            self.remove_closed_tab(0)

    def focus_tab(self, tab_nr):
        """
        Selects the playlist notebook tab tab_nr, and gives it the keyboard
        focus.
        """
        if tab_nr < self.get_n_pages():
            self.set_current_page(tab_nr)
            self.get_current_tab().focus()

    def select_next_tab(self):
        """
        Selects the previous playlist notebook tab, warping around if the
        first page is currently displayed.
        """
        tab_nr = self.get_current_page()
        tab_nr += 1
        tab_nr %= self.get_n_pages()
        self.set_current_page(tab_nr)

    def select_prev_tab(self):
        """
        Selects the next playlist notebook tab, warping around if the last
        page is currently displayed.
        """
        tab_nr = self.get_current_page()
        tab_nr -= 1
        tab_nr %= self.get_n_pages()
        self.set_current_page(tab_nr)

    def on_option_set(self, event, settings, option):
        """
        Updates appearance on setting change
        """
        if option == 'gui/show_tabbar':
            show_tabbar = settings.get_option(option, True)

            if not show_tabbar and self.get_n_pages() > 1:
                show_tabbar = True

            self.set_show_tabs(show_tabbar)

        if option == 'gui/tab_placement':
            tab_placement = settings.get_option(option, 'top')
            self.set_tab_pos(self.tab_placement_map[tab_placement])


class PlaylistContainer(Gtk.Box):
    """
    Contains two playlist notebooks that can contain playlists.
    Playlists can be moved between the two notebooks.

    TODO: Does it make sense to support more than two notebooks?
    I think with this implementation it does not -- we would need to
    move to a different UI design that allowed arbitrary placement
    of UI elements if that was the case.
    """

    def __init__(self, manager_name, player):
        Gtk.Box.__init__(self)

        self.notebooks: List[PlaylistNotebook] = []
        self.notebooks.append(
            PlaylistNotebook(manager_name, player, '<Primary><Shift>t')
        )
        self.notebooks.append(
            PlaylistNotebook(manager_name + '2', player, '<Primary><Alt>t')
        )

        self.notebooks[1].set_add_tab_on_empty(False)

        # add notebooks to self
        self.pack_start(self.notebooks[0], True, True, 0)

        # setup the paned window for separate views
        self.paned = Gtk.Paned(orientation=Gtk.Orientation.VERTICAL)
        self.paned.pack2(self.notebooks[1], True, True)

        # setup queue page
        self.queuepage = QueuePage(self, player)
        self.queuetab = NotebookTab(None, self.queuepage)
        if len(player.queue) > 0:
            self.show_queue()

        # ensure default notebook always has a tab in it
        if self.notebooks[0].get_n_pages() == 0:
            self.notebooks[0].add_default_tab()

        # menu item
        item = menu.simple_menu_item(
            'move-tab',
            [],
            _('_Move to Other View'),
            None,
            lambda w, n, p, c: self._move_tab(p.tab),
            condition_fn=lambda n, p, c: True
            if p.tab.notebook in self.notebooks
            else False,
        )
        providers.register('playlist-tab-context-menu', item)
        providers.register('queue-tab-context', item)

        # connect events
        for notebook in self.notebooks:
            notebook.connect('page-reordered', self.on_page_reordered)
            notebook.connect_after(
                'page-removed', lambda *a: self._update_notebook_display()
            )

        self._update_notebook_display()

    def _move_tab(self, tab):
        if tab.notebook is self.notebooks[0]:
            src, dst = (0, 1)
        else:
            src, dst = (1, 0)

        # don't put this notebook in the 'recently closed tabs' list
        self.notebooks[src]._moving_tab = True
        self.notebooks[src].remove_tab(tab)
        self.notebooks[src]._moving_tab = False

        self.notebooks[dst].add_tab(tab, tab.page)

        # remember where the user moved the queue
        if tab.page is self.queuepage:
            settings.set_option('gui/queue_notebook_num', dst)

        self._update_notebook_display()

    def _update_notebook_display(self):
        pane_installed = self.paned.get_parent() is not None

        if self.notebooks[1].get_n_pages() != 0:
            if not pane_installed:
                parent = self.notebooks[0].get_parent()
                parent.remove(self.notebooks[0])

                self.paned.pack1(self.notebooks[0], True, True)
                self.pack_start(self.paned, True, True, 0)
        else:
            if pane_installed:
                parent = self.notebooks[0].get_parent()
                parent.remove(self.notebooks[0])

                self.remove(self.paned)
                self.pack_start(self.notebooks[0], True, True, 0)

        self.show_all()

    def create_new_playlist(self):
        """
        Create a new tab in the primary notebook containing a blank
        playlist. The tab will be automatically given a unique name.
        """
        return self.notebooks[0].create_new_playlist()

    def create_tab_from_playlist(self, pl):
        """
        Create a tab that will contain the passed-in playlist

        :param playlist: The playlist to create tab from
        :type playlist: :class:`xl.playlist.Playlist`
        """
        return self.notebooks[0].create_tab_from_playlist(pl)

    def get_current_notebook(self):
        """
        Returns the last focused notebook, or the
        primary notebook
        """
        if self.paned.get_parent() is not None:
            focus = self.paned.get_focus_child()
            if focus is not None:
                return focus
        return self.notebooks[0]

    def get_current_tab(self):
        """
        Returns the currently showing tab on the current notebook
        """
        notebook = self.get_current_notebook()
        return notebook.get_current_tab()

    def focus(self):
        """
        Gives keyboard focus to the currently selected tab
        """
        self.get_current_tab().focus()

    def on_page_reordered(self, notebook, child, page_number):
        if (
            self.queuepage.tab.notebook is notebook
            and notebook.page_num(self.queuepage) != 0
        ):
            notebook.reorder_child(self.queuepage, 0)

    def save_current_tabs(self):
        """
        Saves the open tabs
        """
        for notebook in self.notebooks:
            notebook.save_current_tabs()

    def show_queue(self, switch=True):
        """
        Shows the queue page in the last notebook that
        the queue was located.

        :param switch: If True, switch focus to the queue page
        """
        if self.queuepage.tab.notebook is None:
            # ensure the queue is restored in the last place the user had it
            n = settings.get_option('gui/queue_notebook_num', 0)
            self.notebooks[n].add_tab(self.queuetab, self.queuepage, position=0)
        if switch:
            # should always be 0, but doesn't hurt to be safe...
            qnotebook = self.queuepage.tab.notebook
            qnotebook.set_current_page(qnotebook.page_num(self.queuepage))

        self._update_notebook_display()

    def show_current_track(self):
        """
        Tries to find the currently playing track
        and selects it and its containing tab page
        """
        for notebook in self.notebooks:
            if notebook.show_current_track():
                break



# Copyright (C) 2009-2010 Aren Olson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import os

from xl import settings, transcoder
from xl.nls import gettext as _
from xlgui.preferences import widgets

name = _("CD")
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "cdprefs_pane.ui")

FORMAT_WIDGET = None

# TODO: allow setting cddb server?


class OutputFormatPreference(widgets.ComboPreference):
    name = 'cd_import/format'


class OutputQualityPreference(widgets.ComboPreference, widgets.Conditional):
    name = 'cd_import/quality'
    condition_preference_name = 'cd_import/format'

    def __init__(self, preferences, widget):
        widgets.ComboPreference.__init__(self, preferences, widget)
        widgets.Conditional.__init__(self)
        self.format = settings.get_option("cd_import/format", None)
        self.default = settings.get_option("cd_import/quality", None)

    def on_check_condition(self):
        """
        Specifies the condition to meet

        :returns: Whether the condition is met or not
        :rtype: bool
        """
        model = self.widget.get_model()
        if not model:  # happens if preferences window is shut down on close
            return False

        curiter = self.condition_widget.get_active_iter()
        format = self.condition_widget.get_model().get_value(curiter, 0)
        formatinfo = transcoder.FORMATS[format]
        if self.format != format:
            self.format = format
            default = formatinfo['default']

            if self.default != default:
                self.default = default  # raw value

        default_title = formatinfo['kbs_steps'][
            formatinfo['raw_steps'].index(self.default)
        ]
        active_iter = self.widget.get_active_iter()

        if active_iter is not None:
            active_title = float(model.get_value(active_iter, 1))
        else:
            active_title = default_title

        self.widget.set_model(None)
        model.clear()

        steps = zip(formatinfo['raw_steps'], formatinfo['kbs_steps'])

        for item, title in steps:
            model.append([item, str(title)])

        self.widget.set_model(model)

        if active_title not in formatinfo['kbs_steps']:
            active_title = default_title

        index = formatinfo['kbs_steps'].index(active_title)
        self.widget.set_active(index)

        return True


class OutputPathPreference(widgets.ComboEntryPreference):
    name = 'cd_import/outpath'
    completion_items = {
        '$tracknumber': _('Track number'),
        '$title': _('Title'),
        '$artist': _('Artist'),
        '$composer': _('Composer'),
        '$album': _('Album'),
        '$__length': _('Length'),
        '$discnumber': _('Disc number'),
        '$__rating': _('Rating'),
        '$date': _('Date'),
        '$genre': _('Genre'),
        '$bitrate': _('Bitrate'),
        '$__loc': _('Location'),
        '$filename': _('Filename'),
        '$__playcount': _('Play count'),
        '$__last_played': _('Last played'),
        '$bpm': _('BPM'),
    }
    preset_items = ["%s/$artist/$album/$tracknumber - $title" % os.getenv("HOME")]
    default = "%s/$artist/$album/$tracknumber - $title" % os.getenv("HOME")


# vim: et sts=4 sw=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gtk

from xl.nls import gettext as _
from xl import main, playlist, event
from xlgui.widgets import dialogs

from xlgui.widgets.playlist import PlaylistPageBase, PlaylistView


class QueuePage(PlaylistPageBase):
    def __init__(self, container, player):
        PlaylistPageBase.__init__(self)
        self.plcontainer = container
        self.player = player
        self.playlist = player.queue  # a queue is a playlist object...

        self.swindow = Gtk.ScrolledWindow()
        self.swindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.pack_start(self.swindow, True, True, 0)

        self.view = PlaylistView(self.player.queue, self.player)
        self.view.dragdrop_copyonly = True
        self.swindow.add(self.view)

        event.add_ui_callback(
            self.on_length_changed,
            'playlist_current_position_changed',
            self.player.queue,
        )
        event.add_ui_callback(
            self.on_length_changed, "playlist_tracks_added", self.player.queue
        )
        event.add_ui_callback(
            self.on_length_changed, "playlist_tracks_removed", self.player.queue
        )

        self.show_all()

    def on_length_changed(self, *args):
        self.name_changed()
        if len(self.player.queue) == 0:
            self.tab.set_closable(True)
        else:
            self.plcontainer.show_queue(switch=False)
            self.tab.set_closable(False)

    ## NotebookPage API ##

    def focus(self):
        self.view.grab_focus()

    def get_page_name(self):
        qlen = self.player.queue.queue_length()
        if qlen == -1:
            return _("Queue")
        else:
            return _("Queue (%d)") % qlen

    def set_tab(self, tab):
        super(QueuePage, self).set_tab(tab)
        tab.set_closable(not self.do_closing())

    def do_closing(self):
        """
        Allows closing only if the queue is empty
        """
        return len(self.player.queue) != 0

    ## End NotebookPage ##

    def on_saveas(self):
        exaile = main.exaile()
        name = dialogs.ask_for_playlist_name(exaile.gui.main.window, exaile.playlists)
        if name is not None:
            pl = playlist.Playlist(name, self.playlist[:])
            exaile.playlists.save_playlist(pl)
            self.plcontainer.create_tab_from_playlist(pl)


# vim: et sw=4 st=4
